#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
 
# 
# Transform port declarations.
# 
# We transform
# 
#     <in_porttype> pi
#     
# to
# 
#     integer pi
#     
# and if pi is not passed as a parameter, we call
#     
#     call CFallocate_port(pi)
#     
# We transform
# 
#     <out_porttype> po
#     
# to
# 
#     integer po
#     
# and flag an error if if po was not passed as a parameter and does 
# not later occur in a couple statement.
# 
# 
# In the retyping of port variables, we retain the original type as 
# attribute 'originalType'. From this we can lookup the typing 
# information we need to check for correctness of send/receive calls.
#

sub xformPorts
{
    local($var, $type, $real_type, $name, $flavor, $isParallel);

    ($flavor, $isParallel) = split(/$;/, $Programs{$CurrentProgram});

    print "xformPorts on $CurrentProgram ispar=$isParallel flavor=$flavor\n" if $opt_v;

    #
    # We generate port initialization code for common block variables
    # in the main subprogram and in process subprograms.
    #
    $needCommonInit = ($isParallel || $flavor eq 'program');
	
    while (($name, *array) = each(%'Commons))
    {
	for $var (@array)
	{
	    $type = &vardict'getType($var);
	    
	    if ($type eq 'in_port' || $type eq 'out_port')
	    {
		&xformPort($var, $type, $needCommonInit);
	    }
	}
    }

    for $var (@Locals, @Args)
    {
	$type = &vardict'getType($var);

	if ($type eq 'in_port' || $type eq 'out_port')
	{
	    &xformPort($var, $type, 0);
	}
    }
}

sub xformPort
{
    local($var, $type, $needCommonInit) = @_;
    local($dos, $ends, $dim, $subscr, $indent, $alias, $call);
    local(@dims, $len_expr, $flavor, $isCommon, $isArg);

    ($flavor = $type) =~ s/_//;

    $isCommon = &isCommon($var);
    $isArg = &isArg($var);

#    print "xformInPort: var='$var' type='$type' iscommon='$isCommon' needCommonInit=$needCommonInit isArg=$isArg\n";

    &vardict'setOriginalType($var, $type);
    &vardict'setType($var, 'integer');

    #
    # We need to generate the port init code for either variables which
    # are neither arguments nor common block variables, or common
    # block variables for which $needCommonInit is true.
    #
    if ((! ($isArg || $isCommon)) ||
	($isCommon && $needCommonInit))
    {
	local($vexp, $gflav);

#	print "Initializing $var\n";
	if ($opt_g)
	{
	    $vexp = ", '$var'";
	    $gflav = "_debug";
	}

	if (($dim = &vardict'getDimension($var)) ne '')
	{
	    ($len_expr = $dim) =~ s/\|/ * /g;
	}
	else
	{
	    $len_expr = "1";
	}
	&addInitCode("call fm_init_$flavor$gflav($var, $len_expr $vexp)");
    }
}

sub portMessageSizeDeleted
{
    local($name) = @_;
    local($contents, $c);

    $contents = &vardict'get($name, 'contents');
    if ($contents eq '')
    {
	&fatal("Contents not defined for $name");
    }

    $size = 0;
    for $c (split(/\+/, $contents))
    {
	$size += &typeSize($c);
    }
#    print "Got size of message '$name' '$contents' = $size\n";
    return $size;
}
	    
1;
