#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
$FH = "Input0000";
@InputFileStack = ();

&load_file('ftypes.ph');

sub readfile
{
    local($file, $file_name) = @_;
    local($_);

    $InputFile = $FH++;

    &set_input_file($file_name);

    if (!open($InputFile, "<$file"))
    {
	&fatal("Cannot open file $file: $!");
    }

    print "Reading input file $file\n" if $opt_v;

    while (1)
    {
	$_ = &getline;
	last unless defined($_);

	&debug'print(5, "Line: <$_>\n");
	
	next if /^[cC]/;
	next if /^\s*$/;

	$rc = &parse($_ . "\n", $tree, $item_type, $lineno, $failed_stmt);

	if (($rc != 0 || $tree == 0) && $failed_stmt ne '')
	{
	    &error("$file_name:$lineno: Syntax error in $failed_stmt construct");
	    next;
	}

	if ($rc != 0 || $tree == 0)
	{
	    &error("$file_name:$lineno: Syntax error at top level");
	    next;
	}

	$type = &tree_type_name($tree);

#	print "Parsed type $type item_type $item_type\n";

	if ($item_type == &PT_INCLUDE)
	{
	    $file = &tree_string_value(&tree_kid($tree, 0));
#	    print "Got include '$file'\n";
	    push(@includes, $file);
	    &readfile($file);
	}
	elsif ($item_type == &PT_SUBPROGRAM)
	{
#	    print "Got subprogram\n";
#	    &dump_parse_tree($tree);

	    &startSubprogram($tree);
	    &readSubprogram;
	}
	else
	{
	    &error($tree, "Syntax error at top level");
	    next;
	}
    }
}

sub startSubprogram
{
    local($tree) = @_;
    local($name, @args, $type);

    $type = &tree_type_name($tree);
#    &dump_parse_tree($tree);
    $name = &tree_string_value(&tree_kid($tree, 0));

    if ($Programs{$name})
    {
	&error($tree, "Subprogram $name already defined");
    }

    if ($type eq "process")
    {
	@args = &tree_list_to_perl_list(&tree_kid($tree,1));
	&startSubroutine(1, $name, @args);
    }
    elsif ($type eq "subroutine")
    {
	@args = &tree_list_to_perl_list(&tree_kid($tree,1));

	&startSubroutine(0, $name, @args);
    }
    elsif ($type eq "function")
    {
	@args = &tree_list_to_perl_list(&tree_kid($tree,1));
	&startFunction(0, &tree_kid($tree, 2), $name, @args);
    }
    elsif ($type eq "program")
    {
	@args = &tree_list_to_perl_list(&tree_kid($tree,1));
	&startProgram($name, @args);
    }
    $ProgramTree{$name} = $tree;
}

		   
sub startSubroutine
{
    local($isParallel, $name, @args) = @_;

    &debug'print("startSubroutine: '$isParallel' '$name' '@args'\n");
    
    &declareSubprogram('subroutine', $name, $isParallel);

    &declareProgramArg(@args);
}

sub startProgram
{
    local($name, @args) = @_;

    &debug'print("startProgram: '$name' '@args'\n");
    
    &declareSubprogram('program', $name, undef);

    &declareProgramArg(@args);

    if (!$opt_nexus && !$opt_proc)
    {
	&addInitCode("call CFinit($debugLevel)");
    }

    $HaveMainProgram = 1;
}

sub startFunction
{
    local($isParallel, $rettype, $name, @args) = @_;

    if (&tree_num_kids($rettype) == 0)
    {
	#
	# This is an implicitly typed function
	#
	print "Implict function $name\n";
	local($imptype, $t, $l);
	$l = substr($name, 0, 1);
	$imptype = $ImplicitTypes{$l};
	if ($imptype)
	{
	    $t = &new_string_node($imptype);
	    &add_tree_kid($rettype, $t);
	}
	else
	{
	    &error($rettype,
		   "Cannot find implicit type for function $name");
	}
    }

    &debug'print("startFunction: '$name' '@args' '$rettype'\n");
    &declareSubprogram('function', $name, $isParallel, $rettype);

    &declareProgramArg(@args);
}

sub readSubprogram
{
    local($_, $i, $k);
    
    while (1)
    {
	$_ = &getline;

	last if !defined($_);

#	print "readSubprogram gets '$_'\n";

	next if /^[cC]/;
	next if /^\s*$/;

	$rc = &parse($_ . "\n", $tree, $item_type, $lineno, $failed_stmt);

	if (($rc != 0 || $tree == 0) && $failed_stmt ne '')
	{
	    &error("$file_name:$lineno: Syntax error in $failed_stmt construct");
	    next;
	}

	if ($rc == 0 && $tree != 0)
	{
	    if ($item_type == &PT_DECLARATION)
	    {
	        #
		# all the declarations are returned as lists
	        #
		for ($i = 0; $i < &tree_num_kids($tree); $i++)
		{
		    $k = &tree_kid($tree, $i);
		    &addRawCode("DECL:$k");
		}
	    }
	    elsif ($item_type == &PT_INCLUDE)
	    {
		&doInclude(&tree_string_value(&tree_kid($tree, 0)));
	    }
	    elsif ($item_type == &PT_EXECUTABLE)
	    {
		&addRawCode(join("\002", $_, "EXEC:$tree"));
	    }
	    else
	    {
		&addRawCode($_);
	    }
	}
	elsif (/^\s*equivalence/)
	{
	    &addRawCode("EQUIV:$_");
	}
	else
	{
	    &addRawCode($_);
	}
	last if /^\s*end\s*$/i;
    }
    #
    # Drop the 'end' line
    #
    pop(@RawCode);
}

sub doInclude
{
    local($file) = @_;

#    print "Got include '$file'\n";
    push(@includes, $file);

    push(@InputFileStack, $InputFile);
    push(@LA, $lookahead);
    undef($lookahead);
    $InputFile = $FH++;

    if (!open($InputFile, "<$file"))
    {
	&fatal("Cannot open include file $file: $!");
    }

    &readSubprogram;

    close($InputFile);
    $InputFile = pop(@InputFileStack);
    $lookahead = pop(@LA);
}    

1;
