#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
#
# UserTypes is a global, so that we can generate the wrappers for a file.
#
    
%UserTypes = ();

#
# PortDict and PortDictRev are also global
#

%PortDict = ();

%StateVariables = ('RawCode', 'array',
		   'InitCode', 'array',
		   'CleanupCode', 'array',
		   'DeclCode', 'array',
		   'BodyCode', 'array',
		   'Commons', 'assoc',
		   'Locals', 'array',
		   'Parameters', 'assoc',
		   'Args', 'array',
		   'Order', 'array',
		   'ImplicitTypes', 'assoc',
		   'ImplicitStatementList', 'array',
		   'NoImplicitTypes', 'scalar',
		   'CoupledOutputPorts', 'assoc',
		   'VarDict', 'varDict');

@DumpOrder = ('Args',
	      'Parameters',
	      'Locals',
	      'Order',
	      'Commons',
	      'RawCode',
	      'DeclCode',
	      'InitCode',
	      'CleanupCode',
	      'BodyCode',
#	      'ImplicitTypes',
	      'ImplicitStatementList',
#	      'NoImplicitTypes',
	      'VarDict');

%StateVariableInitializers = ('array', 'newArray',
			      'varDict', 'newAssoc',
			      'scalar', 'newScalar',
			      'assoc', 'newAssoc');

%StateVariableDumpers = ('array', 'dumpArray',
			 'assoc', 'dumpAssoc',
			 'assoc', 'dumpSortedAssoc',
			 'scalar', 'dumpScalar',
			 'varDict', 'dumpVarDict');

sub dumpState
{
    local($name, $type, $func);

    for $name (@DumpOrder)
    {
	$type = $StateVariables{$name};
	print "State variable $name:\n";
	$func = $StateVariableDumpers{$type};
	$func eq '' && &fatal("dumpState: type $type has no dump function\n");
	&$func($name);
    }
}

sub dumpVarDict
{
    local(*assoc) = @_;
    local($k, $v, $a, $pair, $attr, $val);

    for $k (sort keys %assoc)
    {
	$v = $assoc{$k};
	print "    $k:\n";
	for $pair (sort split(/$;/, $v))
	{
	    ($attr, $val) = split(/\001/, $pair);
	    print "\t$attr = $val\n";
	}
    }
}

sub dumpScalar
{
    local(*scalar) = @_;

    print "\t$scalar\n";
}

sub dumpArray
{
    local(*array) = @_;

    for (@array)
    {
	print "\t$_\n";
    }
}
	
sub dumpAssoc
{
    local(*assoc) = @_;
    local($k, $v, $a);

    while (($k, $v) = each(%assoc))
    {
	if ($v =~ /ArrayGensym/)
	{
	    local(*array) = $v;
	    
	    print "\t$k: \n";
	    for $a (@array)
	    {
		print "\t    $a\n";
	    }
	}
	else
	{
	    print "\t$k: $v\n";
	}
    }
}

sub dumpSortedAssoc
{
    local(*assoc) = @_;
    local($k, $v, $a);

    for $k (sort keys %assoc)
    {
	$v = $assoc{$k};
	if ($v =~ /ArrayGensym/)
	{
	    local(*array) = $v;
	    
	    print "\t$k: \n";
	    for $a (@array)
	    {
		print "\t    $a\n";
	    }
	}
	else
	{
	    print "\t$k: $v\n";
	}
    }
}

sub setupNewState
{
    local($prog) = @_;
    local($name, $type, $func);

    while (($name, $type) = each(%StateVariables))
    {
	$func = $StateVariableInitializers{$type};
	&fatal("setupNewState: Invalid type $type") if $func eq '';

	eval "\$Prog$name{\$prog} = *$name = &$func";
	&fatal("Eval failed: $@") if $@;
	
    }
}

sub switchToState
{
    local($prog) = @_;
    local($name, $type, $func);

    while (($name, $type) = each(%StateVariables))
    {
	eval "*$name = \$Prog$name{\$prog}";
	&fatal("Eval failed: $@") if $@;
    }
}

sub pushState
{
    local($program) = @_;
#    print "PUSH-1 $program: '@StateStack'\n";

    push(@StateStack, $CurrentProgram);
    &setState($program);
#    print "PUSH-2 $program: '@StateStack'\n";
}

sub setState
{
    local($program) = @_;

    $'CurrentProgram = $program;

    &switchToState($program);
}

sub popState
{
#    print "POP-1: '@StateStack'\n";
    if (@StateStack == 0)
    {
	&fatal("popState: Empty state stack\n");
    }
    $CurrentProgram = pop(@StateStack);
#    print "New state is '$CurrentProgram'\n";
    &switchToState($CurrentProgram);
#    print "POP-2: '@StateStack'\n";
}
    
1;
