/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

#ifdef DEBUG

void FM_INIT_INPORT_DEBUG(f_integer *f_inports,
			  f_integer *n_inports,
			  f_character port_name
			  CHAR_LEN_DECL(port_name_len))
{
    char buf[1024];
    LOCAL_CHAR_LEN_DECL(port_name_len)
    fm_inport_t inport;
    int i;

    CONVERT_CHAR_FTN_TO_C_INPLACE(port_name, port_name_len, buf);

    FM_INIT_INPORT(f_inports, n_inports);

    for (i = 0; i < *n_inports; i++)
    {
	inport = (fm_inport_t) f_inports[i];
	inport->name = (char *) malloc(strlen(buf) + 10);
	MallocCheck(FM_INIT_INPORT_DEBUG(), inport->name, char *);
	sprintf(inport->name, "%s(%d)", buf, i);
    }
}

#endif /* DEBUG */

void FM_INIT_INPORT(f_integer *f_inports, f_integer *n_inports)
{
    fm_inport_t inports, inport;
    int i;

    inports = (fm_inport_t) malloc(sizeof(struct _fm_inport_t) * *n_inports);
    MallocCheck(FM_INIT_INPORT(), inports, fm_inport_t);

    for (i = 0; i < *n_inports; i++)
    {
	inport = &inports[i];
#ifdef DEBUG
	inport->start_magic = FM_INPORT_START_MAGIC;
	inport->end_magic = FM_INPORT_END_MAGIC;
	inport->name = "(unknown)";
#endif
	inport->channel_id = -1;
	inport->channel_conn = NULL;
	inport->outports = NULL;
	inport->n_outports = inport->n_outports_alloc = 0;
	f_inports[i] = (int) inport;
    }
}

#ifdef DEBUG

void FM_INIT_OUTPORT_DEBUG(f_integer *f_outports,
			   f_integer *n_outports,
			   f_character port_name
			   CHAR_LEN_DECL(port_name_len))
{
    char buf[1024];
    fm_outport_t outport;
    int i;
    LOCAL_CHAR_LEN_DECL(port_name_len);

    CONVERT_CHAR_FTN_TO_C_INPLACE(port_name, port_name_len, buf);

    FM_INIT_OUTPORT(f_outports, n_outports);

    for (i = 0; i < *n_outports; i++)
    {
	outport = (fm_outport_t) f_outports[i];
	outport->name = (char *) malloc(strlen(buf) + 10);
	MallocCheck(FM_INIT_OUTPORT_DEBUG(), outport->name, char *);
	sprintf(outport->name, "%s(%d)", buf, i);
    }
}

#endif /* DEBUG */

void FM_INIT_OUTPORT(f_integer *f_outports, f_integer *n_outports)
{
    fm_outport_t outports, outport;
    int i;

    outports = (fm_outport_t)
	malloc(sizeof(struct _fm_outport_t) * *n_outports);
    MallocCheck(FM_INIT_OUTPORT(), outports, fm_outport_t);

    for (i = 0; i < *n_outports; i++)
    {
	outport = &outports[i];
#ifdef DEBUG
	outport->start_magic = FM_OUTPORT_START_MAGIC;
	outport->end_magic = FM_OUTPORT_END_MAGIC;
	outport->name = "(unknown)";
#endif
	outport->channel_id = -1;
	outport->channel_conn = NULL;
	outport->future_home = NULL;
	outport->local_inport = NULL;
	outport->closed = FALSE;
	f_outports[i] = (int) outport;
    }
}


/*
 * FM_CHANNEL() 		Fortran Interface Function
 *
 * Create an association between an inport and an outport. Do not
 * create the actual channel yet. This is implemented as a special case
 * of a merger, where inport->outports is of length 1.
 */
void FM_CHANNEL(f_integer *f_inport, f_integer *f_outport,
		f_integer *iostat, f_integer *have_iostat)
{
    fm_inport_t inport;
    fm_outport_t outport;

    StartSafeRegion;
    
    inport = (fm_inport_t) *f_inport;
    outport = (fm_outport_t) *f_outport;

    FM_INPORT_MAGIC_CHECK(FM_CHANNEL, inport);
    FM_OUTPORT_MAGIC_CHECK(FM_CHANNEL, outport);

    _fm_invalidate_inport(inport);
    _fm_invalidate_outport(outport);

    if (outport->local_inport != (fm_inport_t) NULL)
    {
	yoke_warning("fm_channel_echk(): outport %s already has an inport (%s)\n",
		     _fm_outport_as_string(outport),
		     _fm_inport_as_string(inport));
	SetIOSTAT(FM_IOSTAT_OUTPORT_ALREADY_CONNECTED);

	return;
    }

    if (inport->outports != (fm_outport_reference_t *) NULL)
    {
	SetIOSTAT(FM_IOSTAT_INPORT_ALREADY_CONNECTED);
	return;
    }

#ifdef DEBUG
    if (PortDebug(1))
    {
	yoke_printf("%s: CHANNEL OUT=%s IN=%s\n",
		    _fm_process_name,
		    _fm_outport_as_string(outport),
		    _fm_inport_as_string(inport));
    }
#endif
    
    outport->local_inport = inport;

    inport->outports = (fm_outport_reference_t *) malloc(sizeof(fm_outport_reference_t));

    if (inport->outports == (fm_outport_reference_t *) NULL)
	yoke_fatal("fm_channel_echk(): malloc failed\n");

    inport->n_outports = inport->n_outports_alloc = 1;
    inport->outports[0].local_outport = outport;
    inport->outports[0].remote_outport = (conn_t) NULL;

    outport->inport_index = 0;

    SetIOSTAT(FM_IOSTAT_OK);

    EndSafeRegion;
}


/*
 * FM_MERGER_SETUP()
 *
 * Setup the passed inport as a merger of the given size.
 */
void FM_MERGER_SETUP(f_integer *f_size, f_integer *f_inport,
		     f_integer *slot_idx, f_integer *iostat,
		     f_integer *have_iostat)
{
    fm_inport_t inport;
    int size, i;

    StartSafeRegion;
    
    size = (int) *f_size;
    inport = (fm_inport_t) *f_inport;

    SetIOSTAT(FM_IOSTAT_OK);

    FM_INPORT_MAGIC_CHECK(FM_MERGER_SETUP, inport);
    Assert(size > 0);

    _fm_invalidate_inport(inport);

    if (inport->outports != (fm_outport_reference_t *) NULL)
    {
	SetIOSTAT(FM_IOSTAT_INPORT_ALREADY_CONNECTED);
	return;
    }

#ifdef DEBUG
    if (PortDebug(1))
    {
	yoke_printf("%s: MERGER SETUP IN=%s size=%d\n",
		    _fm_process_name,
		    _fm_inport_as_string(inport), size);
    }
#endif
    
    inport->outports = (fm_outport_reference_t *) malloc(size * sizeof(fm_outport_reference_t));

    if (inport->outports == (fm_outport_reference_t *) NULL)
	yoke_fatal("fm_channel_echk(): malloc failed\n");

    inport->n_outports = inport->n_outports_alloc = size;

    *slot_idx = 0;


    for (i = 0; i < size; i++)
    {
	inport->outports[i].local_outport = (fm_outport_t) NULL;
	inport->outports[i].remote_outport = (conn_t) NULL;
    }
}


/*
 * FM_MERGER_ADD()
 *
 * Add an outport to a merging inport.
 */
void FM_MERGER_ADD(f_integer *f_outport, f_integer *f_inport,
		   f_integer *slot_idx, f_integer *iostat,
		   f_integer *have_iostat)
{
    fm_inport_t inport;
    fm_outport_t outport;

    if (*iostat < 0)
	return;
    
    inport = (fm_inport_t) *f_inport;
    outport = (fm_outport_t) *f_outport;

    FM_INPORT_MAGIC_CHECK(FM_MERGER_ADD, inport);
    FM_OUTPORT_MAGIC_CHECK(FM_MERGER_ADD, outport);

    _fm_invalidate_outport(outport);

    if (outport->local_inport != (fm_inport_t) NULL)
    {
	yoke_warning("fm_channel_echk(): outport %s already has an inport (%s)\n",
		     _fm_outport_as_string(outport),
		     _fm_inport_as_string(inport));
	SetIOSTAT(FM_IOSTAT_OUTPORT_ALREADY_CONNECTED);
	return;
    }

    if (inport->outports == (fm_outport_reference_t *) NULL)
	yoke_fatal("FM_MERGER_ADD() got inport with null outport list\n");

#ifdef DEBUG
    if (PortDebug(1))
    {
	char *s1, *s2;
	s1 = _fm_outport_as_string(outport);
	s2 = _fm_inport_as_string(inport);
	yoke_printf("%s: MERGER ADD OUT=%s IN=%s\n",
		    _fm_process_name,
		    s1, s2);
		     

    }
#endif
    
    outport->local_inport = inport;

    Assert(*slot_idx >= 0 && *slot_idx < inport->n_outports);
    inport->outports[*slot_idx].local_outport = outport;
    inport->outports[*slot_idx].remote_outport = (conn_t) NULL;

    outport->inport_index = *slot_idx;

    (*slot_idx)++;
}


/*
 * FM_MERGER_DONE()
 *
 * Marks the completion of the creation of a merger. Just do some
 * consistency checks.
 */
void FM_MERGER_DONE(f_integer *f_inport, f_integer *slot_idx,
		    f_integer *iostat, f_integer *have_iostat)

{
    fm_inport_t inport;

    inport = (fm_inport_t) *f_inport;

    FM_INPORT_MAGIC_CHECK(FM_MERGER_DONE, inport);
    Assert(*slot_idx == inport->n_outports);

    EndSafeRegion;
}

/*
 * FM_CLOSE_PORT()
 *
 * The ENDCHANNEL statement compiles into a call to this routine. We
 * mark the data structure as closed and send a message to the inport
 * containing our index in his list of ports. (This may not be necessary
 * if the inport is just counting numbers of closed ports, but it might
 * be useful for debugging).
 */
void FM_CLOSE_PORT(f_integer *f_outport,
		   f_integer *iostat, f_integer *have_iostat)
{
    fm_outport_t outport;
    yoke_buffer_t buf;
    conn_t dest;

    StartSafeRegion;

    outport = (fm_outport_t) *f_outport;

    if (outport->closed)
	SetIOSTAT(FM_IOSTAT_CLOSE_TWICE);

    if (outport->local_inport != (fm_inport_t) NULL)
    {
#ifdef DEBUG
	if (PortDebug(1))
	    yoke_printf("%s: Closing port %s locally\n",
			_fm_process_name,
			_fm_outport_as_string(outport));
#endif
	_fm_process_close_channel(outport->channel_id, outport->inport_index);
    }
    else
    {
	conn_t reply;
	
	_fm_n_close_acks_needed++;
	
	dest = outport->channel_conn;

#ifdef DEBUG
	if (PortDebug(1))
	    yoke_printf("%s: Closing port %s remotely\n",
			_fm_process_name,
			_fm_outport_as_string(outport));
#endif
	reply = yoke_self();
	buf = yoke_allocate_buffer(dest,
				   yoke_sizeof_int(dest, 1)+
				   yoke_sizeof_conn(dest, reply), 2);

	yoke_put_int(buf, &outport->inport_index, 1);
	yoke_put_conn(buf, reply);

	yoke_establish(dest);
	yoke_send(dest, buf, FM_CLOSE_TYPE(outport->channel_id));
	yoke_release(dest);
    }
    
    outport->closed = TRUE;

/*
    _fm_invalidate_outport(outport);
*/

    SetIOSTAT(FM_IOSTAT_OK);

    EndSafeRegion;
}    


/*
 * _fm_put_inport()
 *
 * Put the basic port stuff into the buffer. Someone else is worrying
 * about the protocol necessary to do this.
 *
 * Any local outports are converted to remote inports before putting.
 */
void _fm_put_inport(yoke_buffer_t buf, fm_inport_t inport)
{
    fm_outport_reference_t *ref;
    int i;
#ifdef DEBUG
    int len;
#endif

    if (inport->channel_id < 0 || inport->channel_conn == NULL)
	yoke_fatal("Trying to put inport with unallocated channel\n");

#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("%s: _fm_put_inport(): %s\n",
		    _fm_process_name,
		    _fm_inport_as_string(inport));
#endif

    FM_INPORT_MAGIC_CHECK(_fm_put_inport, inport);

#ifdef DEBUG
    len = strlen(inport->name);
    yoke_put_int(buf, &len, 1);
    yoke_put_char(buf, inport->name, len);
#endif
    
    yoke_put_int(buf, &inport->channel_id, 1);
    yoke_put_conn(buf, inport->channel_conn);

    yoke_put_int(buf, &inport->n_outports, 1);
    for (i = 0; i < inport->n_outports; i++)
    {
	ref = &inport->outports[i];
	if (ref->local_outport != (fm_outport_t) NULL)
	{
	    if (ref->local_outport->future_home != (conn_t) NULL)
	    {
		yoke_put_conn(buf, ref->local_outport->future_home);
	    }
	    else
	    {
		yoke_put_conn(buf, yoke_self());
	    }
	}
	else 
	{
	    if (ref->remote_outport == (conn_t) NULL)
		yoke_fatal("Invalid outport reference: has neither local_outport nor remote_inport\n");
	    yoke_put_conn(buf, ref->remote_outport);
	}
    }
}

void _fm_get_inport(yoke_buffer_t buf, fm_inport_t *inportp)
{
    fm_outport_reference_t *ref;
    fm_inport_t inport;
    int i;
#ifdef DEBUG
    int len;
#endif
    fm_inport_t local_inport;
    fm_outport_t *local_outports;
    int n_local_outports;

    inport = *inportp = (fm_inport_t) malloc(sizeof (struct _fm_inport_t));
    if (inport == (fm_inport_t) NULL)
	yoke_fatal("_fm_get_inport(): malloc failed\n");

#ifdef DEBUG
    yoke_get_int(buf, &len, 1);
    inport->name = (char *) malloc(len + 1);
    MallocCheck(_fm_get_inport(), inport->name, char *);
    yoke_get_char(buf, inport->name, len);
    inport->name[len] = '\0';
#endif

    yoke_get_int(buf, &inport->channel_id, 1);
    yoke_get_conn(buf, &inport->channel_conn);
    yoke_get_int(buf, &inport->n_outports, 1);

#ifdef DEBUG
    inport->start_magic = FM_INPORT_START_MAGIC;
    inport->end_magic = FM_INPORT_END_MAGIC;
#endif
    
    if (!yoke_conn_is_local(inport->channel_conn))
	yoke_fatal("Hmm. _fm_get_inport() got a nonlocal inport %s\n",
		   _fm_inport_as_string(inport));

    _fm_pindex_lookup_ports(inport->channel_conn, inport->channel_id, 
			    &local_inport, &local_outports,
			    &n_local_outports);

    inport->outports =
	(fm_outport_reference_t *)
	    malloc(sizeof(fm_outport_reference_t) * inport->n_outports);
    if (inport->outports == (fm_outport_reference_t *) NULL)
	yoke_fatal("_fm_get_inport(): malloc failed\n");
    
    for (i = 0; i < inport->n_outports; i++)
    {
	ref = &inport->outports[i];
	ref->local_outport = (fm_outport_t) NULL;
	yoke_get_conn(buf, &ref->remote_outport);
    }

    for (i = 0; i < n_local_outports; i++)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("Local outport %d=%s\n",
			i,
			_fm_outport_as_string(local_outports[i]));
#endif
	Assert(local_outports[i]->inport_index >= 0);
	Assert(local_outports[i]->inport_index < inport->n_outports);
	ref = &(inport->outports[local_outports[i]->inport_index]);
    }

#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("%s: _fm_get_inport(): %s\n",
		    _fm_process_name,
		    _fm_inport_as_string(inport));
#endif
}
    

int _fm_sizeof_inport(conn_t dest, fm_inport_t inport, int *count)
{
    fm_outport_reference_t *ref;
    int i;
    int size;

    size = yoke_sizeof_int(dest, 2);

    size += yoke_sizeof_conn(dest, inport->channel_conn);

    *count = 3 + inport->n_outports;

#ifdef DEBUG

    size +=
	yoke_sizeof_int(dest, 1) +
	yoke_sizeof_char(dest, strlen(inport->name));

    (*count) += 2;
#endif

    for (i = 0; i < inport->n_outports; i++)
    {
	ref = &inport->outports[i];
	if (ref->local_outport != (fm_outport_t) NULL)
	{
	    if (ref->local_outport->future_home != (conn_t) NULL)
	    {
		size += yoke_sizeof_conn(dest,
					 ref->local_outport->future_home);
	    }
	    else
	    {
		size += yoke_sizeof_conn(dest, yoke_self());
	    }
	}
	else
	{
	    if (ref->remote_outport == (conn_t) NULL)
		yoke_fatal("Invalid outport reference: has neither local_outport nor remote_inport\n");
	    size += yoke_sizeof_conn(dest, ref->remote_outport);
	}
    }
    return size;
}

void _fm_put_outport(yoke_buffer_t buf, fm_outport_t outport)
{
#ifdef DEBUG
    int len;
#endif

#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("%s: _fm_put_outport(): %s\n",
		    _fm_process_name,
		    _fm_outport_as_string(outport));
#endif

    FM_OUTPORT_MAGIC_CHECK(_fm_put_outport, outport);

#ifdef DEBUG
    len = strlen(outport->name);
    yoke_put_int(buf, &len, 1);
    yoke_put_char(buf, outport->name, len);
#endif
    
    yoke_put_int(buf, &outport->channel_id, 1);
    yoke_put_conn(buf, outport->channel_conn);
    yoke_put_int(buf, &outport->inport_index, 1);
    yoke_put_int(buf, &outport->closed, 1);
}

void _fm_get_outport(yoke_buffer_t buf, fm_outport_t *outportp)
{
    fm_outport_t outport;
#ifdef DEBUG
    int len;
#endif
    fm_inport_t local_inport;
    fm_outport_t *local_outports;
    int n_local_outports;
    
    outport = *outportp = (fm_outport_t) malloc(sizeof (struct _fm_outport_t));
    MallocCheck(_fm_get_outport(), outport, fm_outport_t);


#ifdef DEBUG
    outport->start_magic = FM_OUTPORT_START_MAGIC;
    outport->end_magic = FM_OUTPORT_END_MAGIC;
#endif

#ifdef DEBUG
    yoke_get_int(buf, &len, 1);
    outport->name = (char *) malloc(len + 1);
    MallocCheck(_fm_get_outport(), outport->name, char *);
    yoke_get_char(buf, outport->name, len);
    outport->name[len] = '\0';
#endif

    yoke_get_int(buf, &(outport->channel_id), 1);
    yoke_get_conn(buf, &(outport->channel_conn));
    yoke_get_int(buf, &outport->inport_index, 1);
    yoke_get_int(buf, &outport->closed, 1);
    outport->future_home = NULL;
    outport->local_inport= NULL;

    _fm_pindex_lookup_ports(outport->channel_conn, outport->channel_id, 
			    &local_inport, &local_outports,
			    &n_local_outports);

    if (local_inport != (fm_inport_t) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("fm_get_outport(): Assigning local_inport=%s\n",
			_fm_inport_as_string(local_inport));
#endif
	outport->local_inport = local_inport;
    }


#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("%s: _fm_get_outport(): %s\n",
		    _fm_process_name,
		    _fm_outport_as_string(outport));
#endif
}

int _fm_sizeof_outport(conn_t dest, fm_outport_t outport, int *count)
{
    *count = 4;
#ifdef DEBUG
    (*count) += 2;
#endif
    return yoke_sizeof_int(dest, 3) +
#ifdef DEBUG
	yoke_sizeof_int(dest, 1) +
	yoke_sizeof_char(dest, strlen(outport->name)) +
#endif
	yoke_sizeof_conn(dest, outport->channel_conn);
}

void FM_MOVE_INPORT(f_integer *f_from, f_integer *f_to)
{
    fm_inport_t to, from;
    int i;
    fm_outport_t outport;
    fm_outport_reference_t *ref;

    StartSafeRegion;

    to = (fm_inport_t) *f_to;
    from = (fm_inport_t) *f_from;

    FM_INPORT_MAGIC_CHECK(FM_MOVE_INPORT, to);
    FM_INPORT_MAGIC_CHECK(FM_MOVE_INPORT, from);

    if (to->n_outports > 0)
	yoke_fatal("FM_MOVE_INPORT(): destination port %s is already defined\n",
		   PortName(to));

    if (from->n_outports == 0)
	yoke_fatal("FM_MOVE_INPORT(): source port %s is not defined\n",
		   PortName(from));

    to->channel_id = from->channel_id;
    to->channel_conn = from->channel_conn;
    to->outports = from->outports;
    to->n_outports = from->n_outports;
    to->n_outports_alloc = from->n_outports_alloc;

    /*
     * We now need to update the pointers that any locally connected
     * outports have to this inport
     */

    for (i = 0; i < from->n_outports; i++)
    {
	ref = &from->outports[i];
	outport = ref->local_outport;
	if (outport != (fm_outport_t) NULL)
	{
	    FM_OUTPORT_MAGIC_CHECK(FM_MOVE_INPORT, outport);
	    Assert(outport->local_inport != (fm_inport_t) NULL);
	    Assert(outport->local_inport == from);
	    outport->local_inport = to;
	}
    }

    _fm_pindex_remove_inport(from);
    _fm_pindex_add_inport(to);

    _fm_invalidate_inport(from);

    EndSafeRegion;
}
/* FM_MOVE_INPORT() */


void FM_MOVE_OUTPORT(f_integer *f_from, f_integer *f_to)
{
    fm_outport_t to, from;
    fm_inport_t inport;
    fm_outport_reference_t *ref;

    StartSafeRegion;

    to = (fm_outport_t) *f_to;
    from = (fm_outport_t) *f_from;

    FM_OUTPORT_MAGIC_CHECK(FM_MOVE_OUTPORT, to);
    FM_OUTPORT_MAGIC_CHECK(FM_MOVE_OUTPORT, from);

    if (to->channel_id >= 0 || to->local_inport != (fm_inport_t) NULL)
	yoke_fatal("FM_MOVE_OUTPORT(): destination port %s is already defined\n",
		   PortName(to));

    if (from->channel_id < 0 && from->local_inport == (fm_inport_t) NULL)
	yoke_fatal("FM_MOVE_OUTPORT(): source port %s is not defined\n",
		   PortName(from));

    to->channel_id = from->channel_id;
    to->channel_conn = from->channel_conn;
    to->local_inport = from->local_inport;
    to->inport_index = from->inport_index;
    to->future_home = from->future_home;
    to->closed = from->closed;

    /*
     * If our inport is local, update his pointer to me.
     */

    inport = from->local_inport;
    if (inport != (fm_inport_t) NULL)
    {
	FM_INPORT_MAGIC_CHECK(FM_MOVE_OUTPORT, inport);
	Assert(from->inport_index >= 0 || from->inport_index < inport->n_outports);
	/*
	 * Potential problem. Looks like inport doesn't have an outport
	 * list, probably because it got migrated in. The local_inport
	 * field got set when the port came in; perhaps we should do
	 * a scan for local outports as well and build an outport
	 * list. Something to think about...
	 */
	if (inport->outports != (fm_outport_reference_t *) NULL)
	{
	    ref = &(inport->outports[from->inport_index]);
	    Assert(ref->local_outport != (fm_outport_t) NULL);
	    Assert(ref->local_outport == from);
	    ref->local_outport = to;
	}
    }

    _fm_pindex_remove_outport(from);
    _fm_pindex_add_outport(to);
    _fm_invalidate_outport(from);
    EndSafeRegion;
}

