/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */


/* Functions defined in fm_init.c */

extern void FM_INIT();


/* Functions defined in fm_vcomputer.c */

extern conn_t _fm_get_manager(int mgr_idx);
extern conn_t _fm_get_manager_for_conn(conn_t conn);
extern int _fm_n_managers();
extern conn_t _fm_vp_to_manager(int vp);
extern conn_t _fm_vp_to_manager_block(int vp);
extern conn_t _fm_vp_to_manager_cyclic(int vp);
  


/* Functions defined in fm_manager.c */

extern void _fm_manager_init();
extern void _fm_manager_run();
extern void _fm_manager_add_child(int pid, conn_t conn);
extern bool_t _fm_manager_clean_up_children();


/* Functions defined in fm_manager_request.c */

extern conn_t _fm_create_process_request(int node, int procedure_ident,
					 int n_ports, int *port_sizes,
					 int *starting_port_id,
					 int pid, int vc_base,
					 char *subprogram_name,
					 int block_number,
					 int slot,
					 fm_process_block_t pblock);

/* Functions defined in fm_process.c */

extern void _fm_process_init(int vp_id, int vc_offset);
extern void _fm_process_setup_channels(int channel_start, int n_channels,
				       int *channel_sizes);
extern void _fm_process_enqueue_message(int channel_id, yoke_buffer_t buf, bool_t);
extern yoke_buffer_t _fm_process_dequeue_message(int channel_id,
						 bool_t *closed);
extern f_logical _fm_process_channel_is_empty(int channel_id, bool_t *closed);
extern void _fm_process_close_channel(int channel_id, int inport_idx);
extern void FM_SET_VC_SIZE(f_integer *dims, f_integer *n_dims);
extern int _fm_process_acquire_channel(int n_outports);
extern void _fm_process_forward_channel(int channel_id,
					conn_t dest, int dest_id);
extern void _fm_process_print_channel(int channel_id);

extern void _fm_process_set_closed_info(int channel_id,
					int n_outports,
					bool_t *closed_outports);
extern void _fm_process_get_closed_info(int channel_id,
					int *n_outports,
					bool_t **closed_outports);
extern void _fm_process_exit(int exit_code);

/* Functions defined in fm_handler.c */

extern void _fm_register_handlers();
extern fm_remote_start_func_t
    _fm_lookup_remote_start(fm_remote_start_func_id_t procedure_ident);
extern char *
    _fm_lookup_remote_start_name(fm_remote_start_func_id_t procedure_ident);
extern char *
    _fm_lookup_process(fm_remote_start_func_id_t procedure_ident);


/* Functions defined in fm_shutdown.c */

extern void FM_SHUTDOWN();
extern void _fm_shutdown_one_manager(conn_t sender);
extern void _fm_handle_shutdown_request();
extern void _fm_handle_abend(yoke_buffer_t buf);


/* Functions defined in fm_channel.c */

extern void FM_INIT_INPORT(f_integer *f_inports, f_integer *n_inports);
extern void FM_INIT_OUTPORT(f_integer *f_outports, f_integer *n_outports);
extern void FM_CHANNEL(f_integer *f_inport,
		       f_integer *f_outport,
		       f_integer *iostat, f_integer *have_iostat);
extern void FM_CLOSE_PORT(f_integer *f_outport,
			  f_integer *iostat, f_integer *have_iostat);
extern void _fm_put_inport(yoke_buffer_t buf, fm_inport_t inport);
extern void _fm_get_inport(yoke_buffer_t buf, fm_inport_t *inportp);
extern int _fm_sizeof_inport(conn_t dest, fm_inport_t inport, int *count);
extern void _fm_put_outport(yoke_buffer_t buf, fm_outport_t outport);
extern void _fm_get_outport(yoke_buffer_t buf, fm_outport_t *outportp);
extern int _fm_sizeof_outport(conn_t dest, fm_outport_t outport, int *count);

/* Functions defined in fm_pblock.c */

extern void FM_SETUP_PROCBLOCK(f_character f_subprogram_name,
			       f_integer *f_block_number,
			       f_integer *f_size,
			       f_integer *procs,
			       f_integer *n_proc_dims,
			       f_integer *f_pblock
			       CHAR_LEN_DECL(subprogram_name_len));
extern void FM_RESET_PROCBLOCK(f_integer *f_pblock);
extern void FM_END_PROCBLOCK(f_integer *f_pblock);
extern int _fm_lookup_destination(int *loc_info, int loc_info_len,
				  fm_process_block_t pblock,
				  int *new_pid, int *vc_offset);
extern void _fm_setup_generic(int proc_ident, fm_copy_func_t copy_func,
			      fm_inport_t inports[], int n_inports,
			      fm_outport_t outports[], int n_outports,
			      int *loc_info, int loc_info_size,
			      fm_process_block_t pblock);
extern void _fm_fillin_outports(fm_process_block_t pblock);
extern void _fm_spawn_generic(yoke_buffer_t argbuf,
			      void *output_args,
			      fm_process_block_t pblock);

extern void _fm_update_pindex(fm_process_block_t pblock);
extern void _fm_handle_process_return_message(conn_t sender, yoke_buffer_t buf);


/* Functions defined in sr.c */

extern void _fm_ensure_inport_has_channel(fm_inport_t inport);
extern void _fm_send_generic(fm_outport_t outport, yoke_buffer_t buf,
			     f_integer *iostat, f_integer *have_iostat);
extern yoke_buffer_t _fm_recv_generic(fm_inport_t inport,
				      f_integer *iostat,
				      f_integer *have_iostat);
extern void _fm_send_generic_debug(fm_outport_t outport, yoke_buffer_t buf,
				   f_integer *iostat,
				   f_integer *have_iostat,
				   f_character f_file, int line
				   CHAR_LEN_DECL(file_len));
extern yoke_buffer_t _fm_recv_generic_debug(fm_inport_t inport,
					    f_integer *iostat,
					    f_integer *have_iostat,
					    f_character f_file, int line
					    CHAR_LEN_DECL(file_len));
extern void FM_PROBE(f_integer *f_inport, f_integer *iostat,
		     f_integer *have_iostat, f_logical *empty);
extern void _fm_handle_messages();
extern void _fm_handle_message(conn_t sender, int type, yoke_buffer_t buf,
			       bool_t process_queue);
#ifdef DEBUG
extern void _fm_do_presend_debug(conn_t dest,
				 fm_inport_migrec_t inports, int n_inports,
				 fm_outport_migrec_t outports, int n_outports,
				 f_character f_file, int line
				 CHAR_LEN_DECL(file_len));
#endif /* DEBUG */
extern void _fm_do_presend(conn_t dest,
			   fm_inport_migrec_t inports, int n_inports,
			   fm_outport_migrec_t outports, int n_outports);
extern void _fm_handle_port_send(conn_t sender,
				 yoke_buffer_t buf, bool_t queueing);

extern void _fm_invalidate_inport(fm_inport_t inport);
extern void _fm_invalidate_outport(fm_outport_t outport);

/* Functions defined in pindex.c */

extern void _fm_pindex_init();
extern void _fm_pindex_add_outport(fm_outport_t outport);
extern void _fm_pindex_add_inport(fm_inport_t inport);
extern void _fm_pindex_remove_outport(fm_outport_t outport);
extern void _fm_pindex_remove_inport(fm_inport_t inport);
extern void _fm_pindex_remove_outport_with_channel(fm_outport_t outport,
						   conn_t channel_conn,
						   int channel_id);
extern void _fm_pindex_remove_inport_with_channel(fm_inport_t inport,
						  conn_t channel_conn,
						  int channel_id);
extern void _fm_pindex_remove_outport_migrec(conn_t conn, int id,
					     fm_outport_migrec_t migrec);
extern void _fm_pindex_remove_inport_migrec(conn_t conn, int id,
					    fm_inport_migrec_t migrec);
extern void _fm_pindex_add_outport_migrec(conn_t conn, int id,
					  fm_outport_migrec_t migrec);
extern void _fm_pindex_add_inport_migrec(conn_t conn, int id,
					 fm_inport_migrec_t migrec);

extern void _fm_pindex_lookup_ports(conn_t conn, int id,
				    fm_inport_t *inport,
				    fm_outport_t **outports,
				    int *n_outports);
extern void _fm_pindex_lookup_migrecs(conn_t conn, int id,
				     fm_inport_migrec_t *inport_migrec,
				     fm_outport_migrec_t **outport_migrecs,
				     int *n_outport_migrecs);
extern void FM_DUMP_PORT_INDEX(char *str, int len);

/* Functions defined in migrate_new.c */

extern void _fm_migrate_inport(fm_inport_t inport,
			       conn_t new_conn, int new_id);
extern void _fm_migrate_outport(fm_outport_t outport);
extern void _fm_do_inport_migration(conn_t sender, yoke_buffer_t buf, bool_t send_ack);
extern void _fm_do_outport_migration(conn_t sender, yoke_buffer_t buf, bool_t send_ack);
extern void _fm_process_permission_req(yoke_buffer_t in_buf,
				       bool_t process_queue);
extern void _fm_signal_conn(char *func, conn_t conn);

/* Functions defined in migrate.c */

extern void _fm_wait_for_migration_acks(int n_acks_wanted,
					fm_process_block_t pblock,
					fm_inport_migrec_t inport_migrecs,
					int n_inport_migrecs,
					fm_outport_migrec_t outport_migrecs,
					int n_outport_migrecs);
#if 0
extern void _fm_do_inport_migration(conn_t sender, yoke_buffer_t buf);
extern void _fm_do_outport_migration(conn_t sender, yoke_buffer_t buf);
extern void _fm_do_outport_migration_guts(conn_t reply_conn,
					  conn_t channel_conn,
					  int channel_id,
					  int slot_index,
					  int outport_index,
					  int inport_index,
					  conn_t new_conn);
#endif
extern void _fm_notify_outports_of_migration(int *n_acks_needed,
					     fm_inport_migrec_t migrec,
					     int slot_idx,
					     int inport_idx);
extern void _fm_notify_inport_of_migration(int current_slot,
					   int *n_acks_wanted,
					   fm_outport_t outport,
					   int outport_idx,
					   conn_t new_loc);
extern void _fm_forward_channels(fm_process_block_entry_t *ent);
extern void _fm_enqueue_migration_request(yoke_buffer_t buf, int type);
extern void _fm_handle_queued_migration_requests();
extern void _fm_do_last_forward_message(int type, yoke_buffer_t buf);

/* Functions defined in debug.c */

extern void _fm_set_iostat(f_integer *iostat, int have_iostat, int error);
#ifdef DEBUG
extern void _fm_set_iostat_debug(f_integer *iostat, int have_iostat, int error,
				 char *file, int line);
#endif

extern void FM_PAUSE(f_integer *f_id, f_character f_msg
		     CHAR_LEN_DECL(f_msg_len));
extern void _fm_pause(int id, char *msg);
extern void _fm_add_sleeppoint(int id, int secs);
extern void _fm_add_pausepoint(int id);
extern void _fm_sizeof_pausepoints(conn_t dest, int *size, int *count);
extern void _fm_put_pausepoints(yoke_buffer_t buf);
extern void _fm_get_pausepoints(yoke_buffer_t buf);
extern char *_fm_outport_as_string(fm_outport_t outport);
extern char *_fm_inport_as_string(fm_inport_t inport);
extern char *_fm_mtype_name(int type);

