/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

static void
process_queued_outport_migrations(fm_migration_queue_entry_t q_head,
				  fm_process_block_t pblock,
				  fm_inport_migrec_t inport_migrecs,
				  int n_inport_migrecs);

void _fm_wait_for_migration_acks(int n_acks_wanted,
				 fm_process_block_t pblock,
				 fm_inport_migrec_t inport_migrecs,
				 int n_inport_migrecs,
				 fm_outport_migrec_t outport_migrecs,
				 int n_outport_migrecs)
{
    yoke_fatal("Aieee! I don't WANT to run _fm_wait_for_migration_acks()\n");
}

#if 0
/*
 * _fm_wait_for_migration_acks()
 *
 * If the current process block performed any migrations, spin
 * waiting for messages; queueing any messages and waiting for
 * the migration ack. Return when we have received all of
 * the migration acks we are waiting for.
 *
 * The interface here has become horribly overloaded due to bad
 * layering: this routine is used both by the process block routines
 * and the s/r routines. Each may have a list of migrecs which need
 * to be passed. In the case of the pblock, we actually have a list
 * of lists (one list per slot in the block).
 *
 * The usage here is that if there is a non-NULL pblock, use the
 * lists there. Otherwise use the lists passed in.
 */
void _fm_wait_for_migration_acks(int n_acks_wanted,
				 fm_process_block_t pblock,
				 fm_inport_migrec_t inport_migrecs,
				 int n_inport_migrecs,
				 fm_outport_migrec_t outport_migrecs,
				 int n_outport_migrecs)
{
    yoke_buffer_t buf;
    conn_t sender;
    int type;
    int ent_index, port_index;
    fm_inport_migrec_t inport_migrec;
    fm_outport_migrec_t outport_migrec;
    int i, n_inport_migrations_pending;
    fm_process_block_entry_t *ent;
    int ent_idx;
    fm_migration_queue_entry_t q_ent;
    fm_migration_queue_entry_t q_head, q_tail;

    q_head = q_tail = (fm_migration_queue_entry_t) NULL;

    n_inport_migrations_pending = 0;

    if (pblock == (fm_process_block_t) NULL)
    {
	for (i = 0; i< n_inport_migrecs; i++)
	{
	    if (inport_migrecs[i].pending)
		n_inport_migrations_pending++;
	}
    }
    else
    {
	for (ent_idx = 0; ent_idx < pblock->size; ent_idx++)
	{
	    ent = &pblock->entries[ent_idx];

	    if (ent->inport_migrecs != (fm_inport_migrec_t) NULL)
	    {
		for (i = 0; i < ent->n_inports; i++)
		{
		    if (ent->inport_migrecs[i].pending)
			n_inport_migrations_pending++;
		}
	    }
	}
    }
    
    /*
     * Receive Loop
     */

    while (n_acks_wanted > 0)
    {
	
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: Waiting for %d migration acks %d inmigs pending\n",
			_fm_process_name,
			n_acks_wanted,
			n_inport_migrations_pending);
#endif
	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type,  _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (FM_IS_SEND(type))
	{
	    ProcessSendMessage(type, buf);
	}
	else if (FM_IS_CLOSE(type))
	{
	    ProcessCloseMessage(type, buf);
	    yoke_free_buffer(buf);
	}
	else
	{
	    switch (type)
	    {
	    case FM_MTYPE_CLOSE_ACK:
		HandleCloseAck(sender, buf);
		break;
		
	    case FM_MTYPE_SEND_PORTS:
		HandlePortSend(sender, buf);
		break;
		    
	    case FM_MTYPE_INPORT_MIGRATE:
		HandleInportMigrate(sender, buf);
		yoke_free_buffer(buf);
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
		if (0 && n_inport_migrations_pending > 0)
		{
		    q_ent = (fm_migration_queue_entry_t)
			malloc(sizeof(struct _fm_migration_queue_entry_t));
		    MallocCheck(wait_for_migration_acks(), q_ent, fm_migration_queue_entry_t);
		    
		    q_ent->type = type;
		    q_ent->buf = buf;
		    q_ent->next = NULL;

		    if (q_head == (fm_migration_queue_entry_t) NULL)
			q_head = q_tail = q_ent;
		    else
		    {
			q_tail->next = q_ent;
			q_tail = q_ent;
		    }
		}
		else
		{
		    HandleOutportMigrate(sender, buf);
		    yoke_free_buffer(buf);
		}
		break;

	    case FM_MTYPE_INPORT_MIGRATE_ACK:
		yoke_get_int(buf, &ent_index, 1);
		yoke_get_int(buf, &port_index, 1);

#ifdef DEBUG
		if (MigrateDebug(2))
		    yoke_printf("%s: Got inport migration ack from %s for entry=%d port=%d\n",
				_fm_process_name,
				yoke_conn_as_string(sender),
				ent_index, port_index);
#endif
		yoke_free_buffer(buf);
		if (pblock != (fm_process_block_t) NULL)
		{
		    fm_process_block_entry_t *ent;
		    Assert(ent_index >= 0 && ent_index < pblock->size);
		    ent = &(pblock->entries[ent_index]);
		    Assert(port_index >= 0 && port_index < ent->n_inports);
		    Assert(ent->inport_migrecs != (fm_inport_migrec_t) NULL);

		    inport_migrec = &(ent->inport_migrecs[port_index]);
		}
		else
		{
		    Assert(port_index >= 0 && port_index < n_inport_migrecs);
		    inport_migrec = &(inport_migrecs[port_index]);
		}

		Assert(inport_migrec->pending);
		Assert(n_inport_migrations_pending > 0);

		_fm_pindex_remove_inport_with_channel(inport_migrec->inport,
						      inport_migrec->old_conn,
						      inport_migrec->old_id);
		_fm_pindex_add_inport(inport_migrec->inport);
		
		inport_migrec->pending = FALSE;

		n_inport_migrations_pending--;
		if (n_inport_migrations_pending == 0)
		{
		    /*
		     * This is kind of gross, but necessary
		     * since we need to be able to find the list
		     * of migrated inports in order to see if
		     * any given outport migration refers to one of them.
		     * If it does, the migration request must be
		     * forwarded on to the new process.
		     */
		    process_queued_outport_migrations(q_head,
						      pblock,
						      inport_migrecs,
						      n_inport_migrecs);
		}
		n_acks_wanted--;
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE_ACK:
		yoke_get_int(buf, &ent_index, 1);
		yoke_get_int(buf, &port_index, 1);
#ifdef DEBUG
		if (MigrateDebug(2))
		    yoke_printf("%s: Got outport migration ack from %s for entry=%d port=%d\n",
				_fm_process_name,
				yoke_conn_as_string(sender),
				ent_index, port_index);
#endif
		yoke_free_buffer(buf);
		if (pblock != (fm_process_block_t) NULL)
		{
		    fm_process_block_entry_t *ent;
		    Assert(ent_index >= 0 && ent_index < pblock->size);
		    ent = &(pblock->entries[ent_index]);
		    Assert(port_index >= 0 && port_index < ent->n_outports);
		    Assert(ent->outport_migrecs != (fm_outport_migrec_t) NULL);

		    outport_migrec = &(ent->outport_migrecs[port_index]);
		}
		else
		{
		    Assert(port_index >= 0 && port_index < n_outport_migrecs);
		    outport_migrec = &(outport_migrecs[port_index]);
		}

		_fm_pindex_remove_outport_with_channel
		    (outport_migrec->outport,
		     outport_migrec->old_conn,
		     outport_migrec->id);
		_fm_pindex_add_outport(outport_migrec->outport);
		
		Assert(outport_migrec->pending);
		
		outport_migrec->pending = FALSE;
		
		n_acks_wanted--;
		break;
		
	    default:
		yoke_fatal("_fm_wait_for_migration_acks(): Dunno what to do with type %d\n", type);
	    }
	}
	yoke_release_transient(sender);
    }
}


static void
process_queued_outport_migrations(fm_migration_queue_entry_t q_head,
				  fm_process_block_t pblock,
				  fm_inport_migrec_t inport_migrecs,
				  int n_inport_migrecs)
{
    fm_migration_queue_entry_t ent, next_ent;
    conn_t reply_conn, channel_conn, new_conn;
    int inport_index, slot_index;
    int channel_id, outport_index;
    int i, ent_index;
    fm_process_block_entry_t *pb_ent;
    bool_t found_migrated_inport;
    fm_inport_migrec_t inport_migrec;
    
    
    for (ent = q_head; ent != (fm_migration_queue_entry_t) NULL;
	 ent = next_ent)
    {
	next_ent = ent->next;

	yoke_get_conn(ent->buf, &reply_conn);
	yoke_get_conn(ent->buf, &channel_conn);
	yoke_get_int(ent->buf, &channel_id, 1);
	yoke_get_int(ent->buf, &slot_index, 1);
	yoke_get_int(ent->buf, &outport_index, 1);
	yoke_get_int(ent->buf, &inport_index, 1);
	yoke_get_conn(ent->buf, &new_conn);

	yoke_free_buffer(ent->buf);

	if (pblock != (fm_process_block_t) NULL)
	{
	    found_migrated_inport = FALSE;
	    for (ent_index = 0; !found_migrated_inport && ent_index < pblock->size; ent_index++)
	    {
		pb_ent = &(pblock->entries[ent_index]);

		if (pb_ent->inport_migrecs == (fm_inport_migrec_t) NULL)
		    continue;
		
		for (i = 0; !found_migrated_inport && i < pb_ent->n_inports; i++)
		{
		    inport_migrec = &(pb_ent->inport_migrecs[i]);
		    if (inport_migrec->old_conn != (conn_t) NULL && 
			yoke_conn_equal(inport_migrec->old_conn, channel_conn) &&
			channel_id == inport_migrec->old_id)
		    {
			found_migrated_inport = TRUE;
		    }
		}
	    }
	}
	else
	{
	    for (i = 0; i < n_inport_migrecs; i++)
	    {
		inport_migrec = &(inport_migrecs[i]);
		if (yoke_conn_equal(inport_migrec->old_conn, channel_conn) &&
		    channel_id == inport_migrec->old_id)
		{
		    found_migrated_inport = TRUE;
		    break;
		}
	    }
	}

	if (found_migrated_inport)
	{
	    conn_t dest;
	    bool_t was_established;
	    yoke_buffer_t buf;
	    
#ifdef DEBUG
	    if (MigrateDebug(1))
		yoke_printf("process_queued_outport_migrations(): Forwarding migration of %s id=%d to %s id=%d\n",
			    yoke_conn_as_string(channel_conn),
			    channel_id,
			    yoke_conn_as_string(inport_migrec->new_conn),
			    inport_migrec->new_id);
#endif
	    dest = inport_migrec->new_conn;
	    buf = yoke_allocate_buffer(dest,
				       yoke_sizeof_conn(dest, reply_conn) +
				       yoke_sizeof_conn(dest, dest) +
				       yoke_sizeof_conn(dest, new_conn) +
				       yoke_sizeof_int(dest, 4),
				       7);
	    yoke_put_conn(buf, reply_conn);
	    yoke_put_conn(buf, dest);
	    yoke_put_int(buf, &inport_migrec->new_id, 1);
	    yoke_put_int(buf, &slot_index, 1);
	    yoke_put_int(buf, &outport_index, 1);
	    yoke_put_int(buf, &inport_index, 1);
	    yoke_put_conn(buf, new_conn);
	    if (!(was_established = yoke_is_established(dest)))
		yoke_establish(dest);
	    yoke_send(dest, buf, FM_MTYPE_OUTPORT_MIGRATE);
	    /*
	     * We need only signal if we weren't already connected.
	     */
	    if (!was_established)
	    {
		yoke_release(dest);
		_fm_signal_conn("process_queued_outport_migrations()", dest);
	    }
	}
	else
	{
	    _fm_do_outport_migration_guts(reply_conn,
					  channel_conn,
					  channel_id,
					  slot_index,
					  outport_index,
					  inport_index,
					  new_conn);
	}

	free(ent);
    }
}
#endif /* 0 */


/*
 * _fm_forward_channels()
 *
 * Forward the channels for each of the inports in this process
 * block entry to the new process. Use the conn_t contained
 * in ent->process since we know it is established.
 */
void _fm_forward_channels(fm_process_block_entry_t *ent)
{
    int i;
    fm_inport_migrec_t migrec;

    for (i = 0; i < ent->n_inports; i++)
    {
	migrec = &(ent->inport_migrecs[i]);

	if (migrec->new_id >= 0)
	{
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("%s: Forwarding channel %d to %s %d\n",
			    _fm_process_name,
			    migrec->old_id,
			    yoke_conn_as_string(migrec->new_conn),
			    migrec->new_id);
#endif
						

	    Assert(migrec->new_conn == ent->process);

	    _fm_process_forward_channel(migrec->old_id,
					migrec->new_conn,
					migrec->new_id);
	}
    }
}


/*
 * _fm_handle_incoming_migrations()
 *
 * This is invoked by the process remote start routine to receive the
 * channels from the creating process.
 */
void _fm_handle_incoming_migrations(conn_t creator, int n_channels)
{
    int n_channels_left;
    yoke_buffer_t buf;
    int type;

    n_channels_left = n_channels;

    while (n_channels_left > 0)
    {
	
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: Waiting for %d channels\n",
			_fm_process_name,
			n_channels_left);
#endif
	buf = yoke_recv(creator, &type);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s creator=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(creator));
#endif

	if (FM_IS_FWD_BUF(type))
	{
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("%s: Got forwarded channel buf from %s cid=%d\n",
			    _fm_process_name,
			    yoke_conn_as_string(creator),
			    FM_CHANNEL_OF_TYPE(type));
#endif
	    ProcessForwardMessage(type, buf);
	}
	else if (FM_IS_FWD_LAST_BUF(type))
	{
	    ProcessLastForwardMessage(type, buf);
	    yoke_free_buffer(buf);
	    n_channels_left--;
	}
	else
	{
	    yoke_fatal("_fm_handle_incoming_migrations(): Dunno what to do with type %d\n", type);
	}
    }
    yoke_release_transient(creator);
} /* _fm_handle_incoming_migrations() */


/*
 * _fm_notify_outports_of_migration()
 *
 * Notify each outport listed in the migrec's inport of the
 * inport's migration.
 *
 * If the outport is local, this is just a local operation.
 *
 * If the outport is remote, send a migration request to the
 * holder of the outport; and a interruption request to the
 * outport holder's manager.
 *
 * The slot_index is the index of the process call being processed
 * in the process block.
 *
 * The inport_index is the index of this inport in the list of inports
 * for the process call.
 *
 */
void _fm_notify_outports_of_migration(int *n_acks_needed,
				      fm_inport_migrec_t migrec,
				      int slot_index,
				      int inport_index)
{
    int i;
    fm_outport_t outport;
    fm_outport_reference_t *ref;
    fm_inport_t inport;
    conn_t reply_conn;
    yoke_buffer_t buf;
    bool_t *closed_outports;
    int n_outports;
    
    inport = migrec->inport;
    FM_INPORT_MAGIC_CHECK(_fm_notify_outports_of_migration, inport);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("%s: _fm_notify_outports_of_migration(): inport is %s\n",
		    _fm_process_name,
		    _fm_inport_as_string(inport));
#endif

    _fm_process_get_closed_info(migrec->old_id, &n_outports, &closed_outports);
    Assert(n_outports == inport->n_outports);

    for (i = 0; i < inport->n_outports; i++)
    {
	ref = &(inport->outports[i]);

	if (closed_outports[i])
	    continue;

	if (ref->local_outport != (fm_outport_t) NULL)
	{
	    outport = ref->local_outport;
	    FM_OUTPORT_MAGIC_CHECK(_fm_notify_outports_of_migration, outport);
	    Assert(outport->local_inport == inport);

	    _fm_pindex_remove_outport(outport);

	    outport->channel_conn = migrec->new_conn;
	    outport->channel_id = migrec->new_id;

/*	    outport->local_inport = (fm_inport_t) NULL;*/

	    _fm_pindex_add_outport(outport);

	    /*
	     * If we're doing this locally, we have to perform
	     * the pindex update that the migration ack code would
	     * have done.
	     */
	    _fm_pindex_remove_inport_with_channel(migrec->inport,
						  migrec->old_conn,
						  migrec->old_id);
	    _fm_pindex_add_inport(migrec->inport);
	}
	else
	{
	    bool_t was_established;
	    
	    Assert(ref->remote_outport != (conn_t) NULL);
	    reply_conn = yoke_self();
	    buf = yoke_allocate_buffer(ref->remote_outport,
			      yoke_sizeof_conn(ref->remote_outport, reply_conn) +
			      yoke_sizeof_conn(ref->remote_outport, migrec->old_conn) +
			      yoke_sizeof_conn(ref->remote_outport, migrec->new_conn) +
			      yoke_sizeof_int(ref->remote_outport, 4), 7);

	    yoke_put_conn(buf, reply_conn);

	    yoke_put_conn(buf, migrec->old_conn);
	    yoke_put_int(buf, &migrec->old_id, 1);

	    yoke_put_conn(buf, migrec->new_conn);
	    yoke_put_int(buf, &migrec->new_id, 1);

	    yoke_put_int(buf, &slot_index, 1);
	    yoke_put_int(buf, &inport_index, 1);

	    /*
	     * Check for established connection in case the
	     * process we're notifying is the one to which we are
	     * in the process of migrating.
	     */
	    was_established = yoke_is_established(ref->remote_outport);

	    if (!was_established)
		yoke_establish(ref->remote_outport);
	    yoke_send(ref->remote_outport, buf, FM_MTYPE_INPORT_MIGRATE);

	    /*
	     * We need only signal if we weren't already connected.
	     */
	    if (!was_established)
	    {
		yoke_release(ref->remote_outport);
		_fm_signal_conn("_fm_notify_outports_of_migration()", ref->remote_outport);
	    }

	    (*n_acks_needed)++;
	}
    }
}

/*
 * _fm_notify_inport_of_migration()
 *
 * Notify the inport associated with this outport of the outport's
 * impending migration.
 *
 * We send a migration request to the holder of the inport; and an
 * interruption request to the outport holder's manager.
 *
 */
void _fm_notify_inport_of_migration(int current_slot,
				    int *n_acks_wanted,
				    fm_outport_t outport,
				    int outport_index,
				    conn_t new_loc)
{
    conn_t reply_conn, dest;
    yoke_buffer_t buf;
    bool_t was_established;

    dest = outport->channel_conn;
    Assert(dest != (conn_t) NULL);
    Assert(new_loc != (conn_t) NULL);
    
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("%s: _fm_notify_inport_of_migration(): outport is %s new_loc=%s\n",
		    _fm_process_name,
		    _fm_outport_as_string(outport),
		    yoke_conn_as_string(new_loc));
#endif

    reply_conn = yoke_self();
    buf = yoke_allocate_buffer(dest,
			       yoke_sizeof_conn(dest, reply_conn) +
			       yoke_sizeof_conn(dest, dest) +
			       yoke_sizeof_conn(dest, new_loc) +
			       yoke_sizeof_int(dest, 4), 7);

    /*
     * Send ack to me
     */
    yoke_put_conn(buf, reply_conn);

    /*
     * This is the channel in question
     */
    yoke_put_conn(buf, dest);
    yoke_put_int(buf, &outport->channel_id, 1);
    yoke_put_int(buf, &current_slot, 1);
    yoke_put_int(buf, &outport_index, 1);
    yoke_put_int(buf, &outport->inport_index, 1);

    /*
     * This is the new location
     */
    yoke_put_conn(buf, new_loc);

    was_established = yoke_is_established(dest);

    if (!was_established)
	yoke_establish(dest);
    yoke_send(dest, buf, FM_MTYPE_OUTPORT_MIGRATE);

    /*
     * We need only signal if we weren't already connected.
     */
    if (!was_established)
    {
	yoke_release(dest);

	_fm_signal_conn("_fm_notify_inport_of_migration()", dest);
    }

    (*n_acks_wanted)++;
}

/*
 * _fm_do_inport_migration()
 *
 * Process an inport migration request. sender and buf are as
 * returned from yoke_recv_transient(). Do not invoke
 * yoke_release_transient() on the conn_t as the sender is
 * responsible for doing that.
 *
 * The basic responsibility of this function is to update the
 * channel_conn and channel_id fields of any outports which refer
 * to the channel contained in the old_conn/old_id fields of
 * the migration request.
 *
 * We need to check if the outports that we need to update are themselves
 * being migrated. If we find that there is an outport migration record
 * in place for the outport we need to update, we just pack up the
 * migration information and forward it on the the outport's new
 * home. If the new process isn't ready to handle the migration request,
 * it will queue it for later processing.
 */
#if 0
void _fm_do_inport_migration(conn_t sender, yoke_buffer_t buf)
{
    conn_t old_conn, new_conn, reply_conn;
    int old_id, new_id;
    int slot_index, inport_index;
    fm_inport_t inport;
    fm_outport_t outport, *outports;
    fm_inport_migrec_t inport_migrec;
    fm_outport_migrec_t outport_migrec, *outport_migrecs;
    int n_outports, n_outport_migrecs, i, j;
    
    yoke_get_conn(buf, &reply_conn);
    yoke_get_conn(buf, &old_conn);
    yoke_get_int(buf, &old_id, 1);
    yoke_get_conn(buf, &new_conn);
    yoke_get_int(buf, &new_id, 1);
    yoke_get_int(buf, &slot_index, 1);
    yoke_get_int(buf, &inport_index, 1);

#ifdef DEBUG
    if (MigrateDebug(2))
    {
	yoke_printf("%s: _fm_do_inport_migration(): Got reply_conn=%s old_conn=%s old_id=%d\n",
		    _fm_process_name,
		    yoke_conn_as_string(reply_conn),
		    yoke_conn_as_string(old_conn), old_id);
	yoke_printf("%s: _fm_do_inport_migration(): Got new_conn=%s new_id=%d\n",
		    _fm_process_name,
		    yoke_conn_as_string(new_conn), new_id);
	yoke_printf("%s: _fm_do_inport_migration(): And slot_index=%d inport_index=%d\n",
		    _fm_process_name,
		    slot_index, inport_index);
    }
#endif

    _fm_pindex_lookup_ports(old_conn, old_id, &inport, &outports, &n_outports);
    _fm_pindex_lookup_migrecs(old_conn, old_id, &inport_migrec, &outport_migrecs, &n_outport_migrecs);

    if (inport != (fm_inport_t) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Got inport %s\n",
			_fm_process_name,
			_fm_inport_as_string(inport));
#endif
    }

#ifdef DEBUG
    if (inport_migrec != (fm_inport_migrec_t) NULL)
    {
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Got inport_migrec old_conn=%s old_id=%d new_conn=%s new_id=%d pending=%d\n",
			_fm_process_name,
			yoke_conn_as_string(inport_migrec->old_conn),
			inport_migrec->old_id,
			yoke_conn_as_string(inport_migrec->new_conn),
			inport_migrec->new_id,
			inport_migrec->pending);
    }
#endif

#ifdef DEBUG
    if (n_outport_migrecs > 0)
    {
	for (i = 0; i < n_outport_migrecs; i++)
	{
	    outport_migrec = outport_migrecs[i];
	    if (MigrateDebug(2))
		yoke_printf("%s: _fm_do_inport_migration(): Got outport_migrec %d old_conn=%s id=%d new_conn=%s pending=%d\n",
			    _fm_process_name, i,
			    yoke_conn_as_string(outport_migrec->old_conn),
			    outport_migrec->id,
			    yoke_conn_as_string(outport_migrec->new_conn),
			    outport_migrec->pending);
	    
	}
    }
#endif

    if (outports != (fm_outport_t *) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Got %d outports\n",
			_fm_process_name,
			n_outports);
#endif
	for (i = 0; i < n_outports; i++)
	{
	    outport = outports[i];

	    /*
	     * Check to see if this outport is in the process
	     * of being migrated itself.
	     */
	    for (j = 0; j < n_outport_migrecs; j++)
	    {
		outport_migrec = outport_migrecs[j];
		if (outport_migrec->outport == outport)
		{
		    conn_t mgr, dest;
		    yoke_buffer_t buf;
		    bool_t was_established;
		    
#ifdef DEBUG
		    if (MigrateDebug(1))
			yoke_printf("Found a migrating outport when doing inport migration pending=%d\n",
				    outport_migrec->pending);
#endif

		    dest = outport_migrec->new_conn;
		    mgr = _fm_get_manager_for_conn(dest);
		    
#ifdef DEBUG
		    if (MigrateDebug(1))
			yoke_printf("forwarding to %s\n", yoke_conn_as_string(dest));
#endif
		    
		    buf = yoke_allocate_buffer(dest,
					       yoke_sizeof_conn(dest, reply_conn) +
					       yoke_sizeof_conn(dest, old_conn) +
					       yoke_sizeof_conn(dest, new_conn) +
					       yoke_sizeof_int(dest, 4), 7);

		    yoke_put_conn(buf, reply_conn);
		    yoke_put_conn(buf, old_conn);
		    yoke_put_int(buf, &old_id, 1);
		    yoke_put_conn(buf, new_conn);
		    yoke_put_int(buf, &new_id, 1);

		    yoke_put_int(buf, &slot_index, 1);
		    yoke_put_int(buf, &inport_index, 1);

		    if (!(was_established = yoke_is_established(dest)))
			yoke_establish(dest);
		    yoke_send(dest, buf, FM_MTYPE_INPORT_MIGRATE);
		    if (!was_established)
			yoke_release(dest);
		    return;
		}
	    }
	    /*
	     * Update an outport
	     */
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("%s: _fm_do_inport_migration(): Got outport %d: %s\n",
			    _fm_process_name,
			    i,
			    _fm_outport_as_string(outport));
#endif

	    Assert(yoke_conn_equal(outport->channel_conn, old_conn));
	    Assert(outport->channel_id == old_id);

	    _fm_pindex_remove_outport(outport);
	    outport->channel_conn = new_conn;
	    outport->channel_id = new_id;
	    _fm_pindex_add_outport(outport);
	    
	}
    }

    buf = yoke_allocate_buffer(reply_conn, yoke_sizeof_int(reply_conn, 2), 2);
    yoke_put_int(buf, &slot_index, 1);
    yoke_put_int(buf, &inport_index, 1);

    /*
     * No need to StartSafeRegion since we are always and only
     * called from within a safe message loop.
     */
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("%s: _fm_do_inport_migration(): Sending ack to reply=%s\n",
		    _fm_process_name,
		    yoke_conn_as_string(reply_conn));
#endif
    yoke_establish(reply_conn);
    yoke_send(reply_conn, buf, FM_MTYPE_INPORT_MIGRATE_ACK);
    yoke_release(reply_conn);
}

#endif /* 0 */

#if 0

/*
 * _fm_do_outport_migration()
 *
 * Process an outport migration request. sender and buf are as
 * returned from yoke_recv_transient(). Do not invoke
 * yoke_release_transient() on the conn_t as the sender is
 * responsible for doing that.
 */
void _fm_do_outport_migration(conn_t sender, yoke_buffer_t buf)
{
    conn_t reply_conn, channel_conn, new_conn;
    int inport_index, slot_index;
    int channel_id, outport_index;
    
    yoke_get_conn(buf, &reply_conn);
    yoke_get_conn(buf, &channel_conn);
    yoke_get_int(buf, &channel_id, 1);
    yoke_get_int(buf, &slot_index, 1);
    /* This is the slot in the list of outports in the procblock */
    yoke_get_int(buf, &outport_index, 1);
    yoke_get_int(buf, &inport_index, 1);
    yoke_get_conn(buf, &new_conn);

    _fm_do_outport_migration_guts(reply_conn,
				  channel_conn,
				  channel_id,
				  slot_index,
				  outport_index,
				  inport_index,
				  new_conn);
}

void _fm_do_outport_migration_guts(conn_t reply_conn,
				   conn_t channel_conn,
				   int channel_id,
				   int slot_index,
				   int outport_index,
				   int inport_index,
				   conn_t new_conn)
{
    yoke_buffer_t buf;
    fm_inport_t inport;
    fm_outport_t outport, *outports;
    fm_outport_reference_t *ref;
    int n_outports, i;
 
#ifdef DEBUG
    if (MigrateDebug(2))
    {
	yoke_printf("%s: _fm_do_outport_migration(): Got reply_conn=%s channel_conn=%s channel_id=%d slot_index=%d outport_index=%d inport_index=%d\n",
		    _fm_process_name,
		    yoke_conn_as_string(reply_conn),
		    yoke_conn_as_string(channel_conn),
		    channel_id,
		    slot_index,
		    outport_index,
		    inport_index);
	
	yoke_printf("_fm_do_outport_migration(): Got new_conn=%s\n",
		    yoke_conn_as_string(new_conn));
    }
#endif

    _fm_pindex_lookup_ports(channel_conn, channel_id,
			    &inport, &outports, &n_outports);

    if (inport != (fm_inport_t) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("_fm_do_outport_migration(): Got inport %s\n",
			_fm_inport_as_string(inport));
#endif
	/* TODO: IMPORTANT: update pindex */
	Assert(inport_index >= 0 && inport_index < inport->n_outports);
	ref = &(inport->outports[inport_index]);
	Assert(ref->local_outport == (fm_outport_t) NULL);
	Assert(yoke_conn_equal(ref->remote_outport, reply_conn));

	ref->remote_outport = new_conn;
    }
    else
    {
	yoke_fatal("_fm_do_outport_migration(): Didn't find the inport for channel%s id=%d\n",
		   yoke_conn_as_string(channel_conn), channel_id);
    }
    

    if (outports != (fm_outport_t *) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("_fm_do_inport_migration(): Got %d outports\n",
			n_outports);
#endif
	for (i = 0; i < n_outports; i++)
	{
	    outport = outports[i];
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("_fm_do_outport_migration(): Got outport %d: %s\n",
			    i,
			    _fm_outport_as_string(outport));
#endif
	}
    }

    buf = yoke_allocate_buffer(reply_conn, yoke_sizeof_int(reply_conn, 2), 2);
    yoke_put_int(buf, &slot_index, 1);
    yoke_put_int(buf, &outport_index, 1);

    /*
     * No need to StartSafeRegion since we are always and only
     * called from within a safe message loop.
     */
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("_fm_do_outport_migration(): Sending ack to reply=%s\n",
		    yoke_conn_as_string(reply_conn));
#endif
    yoke_establish(reply_conn);
    yoke_send(reply_conn, buf, FM_MTYPE_OUTPORT_MIGRATE_ACK);
    yoke_release(reply_conn);
}
#endif /* 0 */

/*
 * _fm_enqueue_migration_request()
 *
 * Queue up an incoming migration request. This is invoked from
 * process_wait_for_start() when we get a migration request
 * before the process gets started and receives its
 * ports.
 */
void _fm_enqueue_migration_request(yoke_buffer_t buf, int type)
{
    fm_migration_queue_entry_t ent;

    ent = (fm_migration_queue_entry_t)
	malloc(sizeof(struct _fm_migration_queue_entry_t));
    MallocCheck(_fm_enqueue_migration_request, ent, fm_migration_queue_entry_t);

    ent->type = type;
    ent->buf = buf;
    ent->next = NULL;

    if (_fm_migration_q_head == (fm_migration_queue_entry_t) NULL)
	_fm_migration_q_head = _fm_migration_q_tail = ent;
    else
    {
	_fm_migration_q_tail->next = ent;
	_fm_migration_q_tail = ent;
    }
}


/*
 * _fm_handle_queued_migration_requests()
 *
 * Handle any queued up migration requests. This is invoked
 * by the remote start routine just before executing the user
 * code.
 */
void _fm_handle_queued_migration_requests()
{
    fm_migration_queue_entry_t ent, next_ent;

    for (ent = _fm_migration_q_head;
	 ent != (fm_migration_queue_entry_t) NULL;
	 ent = next_ent)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("Handling queued migration request type=%s buf=%x\n",
			_fm_mtype_name(ent->type), ent->buf);
#endif
	if (FM_IS_SEND(ent->type))
	{
	    ProcessSendMessage(ent->type, ent->buf);
	}
	else if (FM_IS_CLOSE(ent->type))
	{
	    ProcessCloseMessage(ent->type, ent->buf);
	}
	else if (FM_IS_FWD_BUF(ent->type))
	{
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("Got forwarded channel buf cid=%d\n",
			    FM_CHANNEL_OF_TYPE(ent->type));
#endif
	    ProcessForwardMessage(ent->type, ent->buf);
	}
	else if (FM_IS_FWD_LAST_BUF(ent->type))
	{
	    ProcessLastForwardMessage(ent->type, ent->buf);
	    yoke_free_buffer(ent->buf);
	}
	else
	{
	    switch (ent->type)
	    {
	    case FM_MTYPE_OUTPORT_PERM_REQ:
		ProcessPermReqNoQueueProcessing(ent->type, ent->buf);
		yoke_free_buffer(ent->buf);
		break;
		
	    case FM_MTYPE_INPORT_MIGRATE:
		_fm_do_inport_migration((conn_t) NULL, ent->buf, TRUE);
		yoke_free_buffer(ent->buf);
		break;
	    
	    case FM_MTYPE_OUTPORT_MIGRATE:
		_fm_do_outport_migration((conn_t) NULL, ent->buf, TRUE);
		yoke_free_buffer(ent->buf);
		break;

	    case FM_MTYPE_INPORT_MIGRATE_NO_ACK:
		_fm_do_inport_migration((conn_t) NULL, ent->buf, FALSE);
		yoke_free_buffer(ent->buf);
		break;
	    
	    case FM_MTYPE_OUTPORT_MIGRATE_NO_ACK:
		_fm_do_outport_migration((conn_t) NULL, ent->buf, FALSE);
		yoke_free_buffer(ent->buf);
		break;

	    default:
		yoke_fatal("_fm_handle_queued_migration_requests(): Invalid buffer type %s %d\n",
			   _fm_mtype_name(ent->type), ent->type);
		
	    }
	}
	next_ent = ent->next;
	free(ent);
    }
    _fm_migration_q_head = _fm_migration_q_tail =
	(fm_migration_queue_entry_t) NULL;
}


/*
 * _fm_do_last_forward_message()
 *
 * Process the last forward message, which contains the list of
 * closed outports for the inport associated with this channel.
 */
void _fm_do_last_forward_message(int type, yoke_buffer_t buf)
{
    int n_outports, n_closed_outports, channel, i, index;
    bool_t *closed_outports;

    channel = FM_CHANNEL_OF_TYPE(type);

    _fm_process_get_closed_info(channel, &n_outports,&closed_outports);

    yoke_get_int(buf, &n_closed_outports, 1);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("Got last forwarded channel buf cid=%d n_closed=%d\n",
		    FM_CHANNEL_OF_TYPE(type),
		    n_closed_outports);
#endif
    for (i = 0; i < n_closed_outports; i++)
    {
	yoke_get_int(buf, &index, 1);
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("inport_index=%d is closed\n", index);
#endif
	closed_outports[index] = TRUE;
    }
    _fm_process_set_closed_info(channel, n_outports, closed_outports);
}

