/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

#ifdef DEBUG

void _fm_do_presend_debug(conn_t dest,
			  fm_inport_migrec_t inports, int n_inports,
			  fm_outport_migrec_t outports, int n_outports,
			  f_character f_file, int line
			  CHAR_LEN_DECL(file_len))
{
    char file[1024];
    LOCAL_CHAR_LEN_DECL(file_len)
    int i;
    fm_inport_t inport;
    fm_outport_t outport;

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_file, file_len, file);
    
    for (i = 0; i < n_inports; i++)
    {
	inport = inports[i].inport;
	if (inport == (fm_inport_t) NULL)
	    yoke_fatal("SEND at %s:%d got null inport\n", file, line);
    
	if (inport->start_magic != FM_INPORT_START_MAGIC ||
	    inport->end_magic != FM_INPORT_END_MAGIC)
	    yoke_fatal("SEND at %s:%d has inport with invalid magic numbers (%x %x)\n",
		       file, line,
		       inport->start_magic, inport->end_magic);
    }

    for (i = 0; i < n_outports; i++)
    {
	outport = outports[i].outport;
	if (outport == (fm_outport_t) NULL)
	    yoke_fatal("SEND at %s:%d got null outport\n", file, line);
    
	if (outport->start_magic != FM_OUTPORT_START_MAGIC ||
	    outport->end_magic != FM_OUTPORT_END_MAGIC)
	    yoke_fatal("SEND at %s:%d has outport with invalid magic numbers (%x %x)\n",
		       file, line,
		       outport->start_magic, outport->end_magic);
    }

    _fm_do_presend(dest, inports, n_inports, outports, n_outports);
}

void _fm_send_generic_debug(fm_outport_t outport, yoke_buffer_t buf,
			    f_integer *iostat, f_integer *have_iostat,
			    f_character f_file, int line
			    CHAR_LEN_DECL(file_len))
{
    char file[1024];
    LOCAL_CHAR_LEN_DECL(file_len)

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_file, file_len, file);

    if (outport == (fm_outport_t) NULL)
	yoke_fatal("SEND at %s:%d got null outport\n", file, line);
    
    if (outport->start_magic != FM_OUTPORT_START_MAGIC ||
	outport->end_magic != FM_OUTPORT_END_MAGIC)
	yoke_fatal("SEND at %s:%d has outport with invalid magic numbers (%x %x)\n",
		   file, line,
		   outport->start_magic, outport->end_magic);


    if (outport->local_inport == (fm_inport_t) NULL &&
	outport->channel_conn == (conn_t) NULL)
yoke_fatal("SEND at %s:%d got on undefined port %s. (Did you pass it to a process?)\n",
	   file, line, PortName(outport));

    _fm_send_generic(outport, buf, iostat,have_iostat);
}

yoke_buffer_t _fm_recv_generic_debug(fm_inport_t inport,
				     f_integer *iostat,
				     f_integer *have_iostat,
				     f_character f_file, int line
				     CHAR_LEN_DECL(file_len))
{
    char file[1024];
    LOCAL_CHAR_LEN_DECL(file_len)

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_file, file_len, file);
    
    if (inport == (fm_inport_t) NULL)
	yoke_fatal("RECEIVE at %s:%d got null inport\n", file, line);
    
    if (inport->start_magic != FM_INPORT_START_MAGIC ||
	inport->end_magic != FM_INPORT_END_MAGIC)
	yoke_fatal("RECEIVE at %s:%d has inport with invalid magic numbers (%x %x)\n",
		   file, line,
		   inport->start_magic, inport->end_magic);

    return _fm_recv_generic(inport, iostat, have_iostat);
}

#endif /* DEBUG */


/*
 * _fm_do_presend()
 *
 * Perform any necessary housekeeping before sending a port over a port.
 *
 * For inports:
 *
 *   Message the destination process for n_inports new channels
 *      (this places the process in a state where it only receives
 *       from this process, thus ensuring message ordering).
 *   Remove the local inport from the port index
 *   Update the fields in the inport data structure
 *   Add the local inport to the port index
 *   Message the holders of the outports of this inport.
 *      (at this point the destination process can receive SEND
 *       messages for the new channel. It will enqueue them to later
 *       be placed in the channel after the forwarded message buffers)
 *   Forward any queued message buffers to the destination process
 *
 * For outports:
 *
 *   If there is no local channel allocated for the port, create
 *       one and update the outport and its inport, and add them to the
 *       port index
 *   Notify the holder of the inport for this outport of the change
 *
 * Wait for the notification acknowledgements to return.
 *
 *
 */
void _fm_do_presend(conn_t dest,
		    fm_inport_migrec_t inports, int n_inports,
		    fm_outport_migrec_t outports, int n_outports)
{
    int i, j;
    fm_inport_t inport;
    fm_outport_t outport;
    int n_acks_needed;
    yoke_buffer_t buf;
    int first_channel;
    fm_inport_migrec_t migrec;
    fm_outport_reference_t *ref;
    int type, id, cid;
    int *port_sizes;

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("_fm_do_presend() dest=%s\n",
		    yoke_conn_as_string(dest));
#endif
    /*
     * Send off the request for channels.
     */

    port_sizes = (int *) malloc(sizeof(int) * n_inports);
    for (i = 0; i < n_inports; i++)
	port_sizes[i] = inports[i].inport->n_outports;

    yoke_establish(dest);
    buf = yoke_allocate_buffer(dest, yoke_sizeof_int(dest, 1 + n_inports),
			       1 + n_inports);
    yoke_put_int(buf, &n_inports, 1);
    yoke_put_int(buf, port_sizes, n_inports);
    yoke_send(dest, buf, FM_MTYPE_SEND_PORTS);

    free(port_sizes);

    n_acks_needed = 0;

    /*
     * While we're waiting, do some consistency checks and port index
     * updates.
     */

    for (i = 0; i < n_inports; i++)
    {
	inport = inports[i].inport;

	FM_INPORT_MAGIC_CHECK(_fm_do_presend, inport);
    }

    /*
     * We can also process our outports while we're waiting for the
     * inport channel request to return. However, we can't assign
     * channels to outports until we've assigned channels to the inports,
     * or at least marked which outports are connected to inports which
     * will be getting channels.
     *
     * TODO: optimize this some more.
     */

#ifdef DEBUG
    for (i = 0; i < n_outports; i++)
    {
	outport = outports[i].outport;

	FM_OUTPORT_MAGIC_CHECK(_fm_do_presend, outport);

	/*
	 * If we have a local inport, do some integrity checks.
	 */
	if (outport->local_inport != (fm_inport_t) NULL)
	{
	    inport = outport->local_inport;
	    FM_INPORT_MAGIC_CHECK(_fm_do_presend, inport);

	    Assert(inport->outports != (fm_outport_reference_t *) NULL);
	    Assert(outport->inport_index >= 0 && outport->inport_index < inport->n_outports);
		    
	    ref = &(inport->outports[outport->inport_index]);
		    
	    Assert(ref->local_outport != (fm_outport_t) NULL);
	    Assert(ref->local_outport == outport);

	    if (outport->channel_id < 0)
	    {
		Assert(inport->channel_id < 0);
	    }
	    else
	    {
		Assert(outport->channel_id == inport->channel_id);
	    }
	}
    }
#endif /* DEBUG */

    /*
     * Receive our new channel ids. This is just passed as
     * the starting index of the first channel.
     */
    buf = yoke_recv(dest, &type);
    if (type != FM_MTYPE_SEND_PORTS_REPLY)
	yoke_fatal("_fm_do_presend(): Expected FM_MTYPE_SEND_PORTS_REPLY, got %s %d\n",
		   _fm_mtype_name(type), type);
    yoke_get_int(buf, &first_channel, 1);
    yoke_free_buffer(buf);

    /*
     * Now we can assign the new channels to the inports.
     */

    for (id = first_channel, i = 0; i < n_inports; i++, id++)
    {
	inport = inports[i].inport;
	FM_INPORT_MAGIC_CHECK(_fm_do_presend, inport);

	if (inport->channel_id < 0)
	{
	    /*
	     * Assign the channel to the inport.
	     *
	     * We don't need to do any migration here. We do update the
	     * port index with entries for the inport and all of its
	     * outports. This is to maintain the invariant that any
	     * port with an assigned channel has a port index entry.
	     */

	    inport->channel_id = id;
	    inport->channel_conn = dest;
	    _fm_pindex_add_inport(inport);

	    for (j = 0; j < inport->n_outports; j++)
	    {
		ref = &(inport->outports[j]);
		outport = ref->local_outport;
		Assert(outport != (fm_outport_t) NULL);
		FM_OUTPORT_MAGIC_CHECK(_fm_do_presend, outport);
	        outport->channel_id = id;
		outport->channel_conn = dest;
		_fm_pindex_add_outport(outport);
	    }
	}
	else
	{
            _fm_migrate_inport(inport, dest, id);
            _fm_pindex_remove_inport(inport);
            inport->channel_conn = dest;
            inport->channel_id = id;
#if 0
	    /*
	     * Do the migration. We create a migrec here
	     * so that we can reuse the notify_outports routine
	     * used in process blocks.
	     */

	    migrec = &(inports[i]);
	     
	    migrec->new_conn = dest;
	    migrec->new_id = id;
	    migrec->old_conn = inport->channel_conn;
	    migrec->old_id = inport->channel_id;

	    _fm_pindex_remove_inport(inport);

	    inport->channel_id = id;
	    inport->channel_conn = dest;

	    _fm_pindex_add_inport(inport);
	    _fm_notify_outports_of_migration(&n_acks_needed, migrec, -1, i);
#endif
	}
    }

    /*
     * Scan the outports looking for any that remain without
     * an assigned channel. Assign them a channel. Notify any
     * remote inports of the move.
     */

    for (i = 0; i < n_outports; i++)
    {
	outport = outports[i].outport;

	if (outport->channel_id < 0)
	{
	    /*
	     * Chase down the inport associated with this outport
	     * and make sure it references back to here.
	     *
	     * Can we ever have an outport with no local inport AND
	     * no channel? The answer was yes during a process block
	     * but I think it is no during a port send.
	     *
	     * Aha! The inport might not be local if the channel
	     * was created for an inport being sent in the same
	     * message. We *do* have to allow for this case here.
	     *
	     * But then again, even if the the inport pointer should
	     * not have been invalidated yet. We'll leave the
	     * inport != NULL assertion in for now.
	     */

	    inport = outport->local_inport;
	    Assert(inport != (fm_inport_t) NULL);
	    if (inport != (fm_inport_t) NULL)
	    {
		FM_INPORT_MAGIC_CHECK(_fm_do_presend, inport);
		Assert(inport->outports != (fm_outport_reference_t *) NULL);
		Assert(outport->inport_index >= 0 && outport->inport_index < inport->n_outports);
		
		ref = &(inport->outports[outport->inport_index]);
		
		Assert(ref->local_outport != (fm_outport_t) NULL);
		Assert(ref->local_outport == outport);
		
		Assert(inport->channel_id < 0);
	    }
	    
	    /*
	     * Allocate a channel and point both ports at it.
	     */
	    
	    cid = _fm_process_acquire_channel(inport->n_outports);
	    
	    if (inport != (fm_inport_t) NULL)
	    {
		inport->channel_id = cid;
		inport->channel_conn = yoke_self();
		
		_fm_pindex_add_inport(inport);
		
#ifdef DEBUG
		if (PortDebug(3))
		    yoke_printf("_fm_do_presend(): Filled in outport %s inport %s cid=%d conn=%x\n",
				_fm_outport_as_string(outport),
				_fm_inport_as_string(inport),
				cid, inport->channel_conn);
#endif
		for (j = 0; j < inport->n_outports; j++)
		{
		    ref = &(inport->outports[j]);
		    
		    Assert(ref->local_outport != (fm_outport_t) NULL);
		    
		    outport = ref->local_outport;
		    
		    Assert(outport->local_inport != (fm_inport_t) NULL);
		    Assert(outport->local_inport == inport);
		    Assert(outport->channel_id < 0);
		    
		    outport->channel_id = cid;
		    outport->channel_conn = inport->channel_conn;
		    
		    _fm_pindex_add_outport(outport);
		}
	    }
	    else
	    {
		/*
		 * The inport isn't local, so we just assign the
		 * channel locally, and add to pindex.
		 */
		
		outport->channel_id = cid;
		outport->channel_conn = yoke_self();
		_fm_pindex_add_outport(outport);
	    }
	}
	else
	{
#if 0
	    outports[i].old_conn = outport->channel_conn;
	    outports[i].new_conn = dest;
	    outports[i].id = outport->channel_id;
#endif
	}

	if (outport->local_inport == (fm_inport_t) NULL)
        {
            outport->future_home = dest;
            _fm_migrate_outport(outport);
#if 0
	    _fm_notify_inport_of_migration(-1, &n_acks_needed,
					   outport, i, dest);
#endif
	}
    }

#if 0
    /*
     * Wait for the acks to come in
     */

    _fm_wait_for_migration_acks(n_acks_needed,
				(fm_process_block_t) NULL,
				inports,
				n_inports,
				outports,
				n_outports);
#endif

    /*
     * Forward any inport channels to the destination process.
     *
     * _fm_process_forward_channel() assumes the conn is established.
     */

    for (i= 0; i < n_inports; i++)
    {
	migrec = &(inports[i]);

	_fm_process_forward_channel(migrec->old_id, migrec->new_conn,
				    migrec->new_id);
    }
}


/*
 * _fm_handle_port_send()
 *
 * Handle an FM_MTYPE_SEND_PORTS message:
 *
 *     Create the requested number of channels, returning the id of the
 *         first one
 *     Loop receiving messages only from the sender until a
 *         SEND is received. We expect to receive only
 *         migration notifications and forwarded channels.
 * 
 */
void _fm_handle_port_send(conn_t sender, yoke_buffer_t buf, bool_t queueing)
{
    int n_channels, channel_start;
    bool_t done;
    int type;
    int *channel_sizes;

    yoke_get_int(buf, &n_channels, 1);
    channel_sizes = (int *) malloc(n_channels * sizeof(int));
    yoke_get_int(buf, channel_sizes, n_channels);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("_fm_handle_port_send(queueing=%c): Port send wants %d channels\n",
                    queueing ? 'T' : 'F', n_channels);
#endif

    channel_start = _fm_next_channel_id;

    _fm_process_setup_channels(channel_start, n_channels, channel_sizes);

    free(channel_sizes);
    
    yoke_free_buffer(buf);

    buf = yoke_allocate_buffer(sender, yoke_sizeof_int(sender, 1), 1);
    yoke_put_int(buf, &channel_start, 1);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_handle_port_send(): replying with channel_start=%d\n",
                    channel_start);
#endif

    yoke_send(sender, buf, FM_MTYPE_SEND_PORTS_REPLY);

    done = FALSE;
    while (!done)
    {
	buf = yoke_recv(sender, &type);

#ifdef DEBUG
    if (MigrateDebug(3))
	yoke_printf("_fm_handle_port_send(): recd type=%s %d\n",
                    _fm_mtype_name(type), type);
#endif
	if (FM_IS_SEND(type))
	{
	    if (queueing)
	    {
		_fm_enqueue_migration_request(buf, type);
	    }
	    else
	    {
		ProcessSendMessage(type, buf);
	    }
            done = 1;
	}
	else if (FM_IS_FWD_BUF(type))
	{
#ifdef DEBUG
	    if (MigrateDebug(2))
		yoke_printf("Got forwarded channel buf from %s cid=%d\n",
			    yoke_conn_as_string(sender),
			    FM_CHANNEL_OF_TYPE(type));
#endif
	    if (queueing)
	    {
		_fm_enqueue_migration_request(buf, type);
	    }
	    else
	    {
		ProcessForwardMessage(type, buf);
	    }
	}
	else if (FM_IS_FWD_LAST_BUF(type))
	{
	    if (queueing)
	    {
		_fm_enqueue_migration_request(buf, type);
	    }
	    else
	    {
		ProcessLastForwardMessage(type, buf);
		yoke_free_buffer(buf);
	    }
	}
	else
	{
	    switch (type)
	    {
	    case FM_MTYPE_INPORT_MIGRATE:
		if (queueing)
		    _fm_enqueue_migration_request(buf, type);
		else
		{
		    HandleInportMigrate(sender, buf);
		    yoke_free_buffer(buf);
		}
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
		if (queueing)
		    _fm_enqueue_migration_request(buf, type);
		else
		{
		    HandleOutportMigrate(sender, buf);
		    yoke_free_buffer(buf);
		}
		break;

	    default:
		yoke_fatal("_fm_handle_port_send(): Don't know what to do with type %s %d\n",
			   _fm_mtype_name(type), type);
	    }
	}

    }
}


/*
 * _fm_ensure_inport_has_channel()
 *
 * If the inport does not have a local channel, create one and notify
 * its outports. This is only called when we know the outports
 * are local (if we have remote outports, then we created a channel
 * during the process of spawning the process who has the remote port).
 */
void _fm_ensure_inport_has_channel(fm_inport_t inport)
{
    fm_outport_t outport;
    fm_outport_reference_t *ref;
    int cid, j;
    conn_t conn;

    if (inport->channel_id >= 0)
	return;

    /*
     * We need to create a channel for this port.
     * Allocate it, update the inport and all of its
     * outports, and add it to the port index.
     */

    FM_INPORT_MAGIC_CHECK(_fm_ensure_inport_has_channel, inport);

    cid = _fm_process_acquire_channel(inport->n_outports);
    conn = yoke_self();
    
    inport->channel_id = cid;
    inport->channel_conn = yoke_self();
    
    _fm_pindex_add_inport(inport);
    
#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("_fm_ensure_inport_has_channel(): allocated channel for inport %s cid=%d conn=%x\n",
		    _fm_inport_as_string(inport),
		    cid, conn);
#endif
    for (j = 0; j < inport->n_outports; j++)
    {
	ref = &(inport->outports[j]);
	
	Assert(ref->local_outport != (fm_outport_t) NULL);
	
	outport = ref->local_outport;
	
	Assert(outport->local_inport != (fm_inport_t) NULL);
	Assert(outport->local_inport == inport);
	Assert(outport->channel_id < 0);
	
	outport->channel_id = cid;
	outport->channel_conn = inport->channel_conn;
	
	_fm_pindex_add_outport(outport);
    }
}

void _fm_send_generic(fm_outport_t outport, yoke_buffer_t buf,
		      f_integer *iostat, f_integer *have_iostat)
{
    FM_OUTPORT_MAGIC_CHECK(_fm_send_generic, outport);
    
#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("Sending to %s\n",
		    _fm_outport_as_string(outport));
#endif

    if (outport->local_inport != (fm_inport_t) NULL)
    {
	Assert(outport->channel_id >= 0);
	_fm_process_enqueue_message(outport->channel_id, buf, FALSE);
    }
    else
    {
	bool_t is_est;
#ifdef DEBUG
	static int time = 1000;
	if (PortDebug(2))
	{
	    _fm_pause(time, "In _fm_send_generic");
	    time++;
	}
#endif

#ifdef DEBUG
	if (outport->channel_conn == (conn_t) NULL)
	    yoke_fatal("Sending on undefined port %s. (Did you pass it to a process?)\n",
		       PortName(outport));
#endif
	
	is_est = yoke_is_established(outport->channel_conn);
	if (!is_est)
	{
	    yoke_establish(outport->channel_conn);
	}
	yoke_send(outport->channel_conn, buf,
		  FM_SEND_TYPE(outport->channel_id));
	yoke_release(outport->channel_conn);
    }

    SetIOSTAT(FM_IOSTAT_OK);

} /* _fm_send_generic() */

yoke_buffer_t _fm_recv_generic(fm_inport_t inport,
			       f_integer *iostat, f_integer *have_iostat)
{
    yoke_buffer_t buf;
    conn_t sender;
    int type;
    bool_t closed;
    int cid, j;

    FM_INPORT_MAGIC_CHECK(_fm_recv_generic, inport);

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("Receiving from %s\n",
		    _fm_inport_as_string(inport));
#endif

    if (inport->channel_id < 0)
    {
	fm_outport_t outport;
	fm_outport_reference_t *ref;
	conn_t conn;
	
	/*
	 * We need to create a channel for this port.
	 * Allocate it, update the inport and all of its
	 * outports, and add it to the port index.
	 */
	
	cid = _fm_process_acquire_channel(inport->n_outports);
	conn = yoke_self();
	
	inport->channel_id = cid;
	inport->channel_conn = yoke_self();
	
	_fm_pindex_add_inport(inport);
	
#ifdef DEBUG
	if (PortDebug(3))
	    yoke_printf("Allocated channel in recv: inport %s cid=%d conn=%x\n",
			_fm_inport_as_string(inport),
			cid, conn);
#endif
	for (j = 0; j < inport->n_outports; j++)
	{
	    ref = &(inport->outports[j]);
	    
	    Assert(ref->local_outport != (fm_outport_t) NULL);
	    
	    outport = ref->local_outport;
	    
	    Assert(outport->local_inport != (fm_inport_t) NULL);
	    Assert(outport->local_inport == inport);
	    Assert(outport->channel_id < 0);
	    
	    outport->channel_id = cid;
	    outport->channel_conn = inport->channel_conn;
	    
	    _fm_pindex_add_outport(outport);
	}
    }

    /*
     * Receive Loop: does not require SafeRegion bracketing
     *               because the compiler-generated receive is bracketed.
     */

    buf = _fm_process_dequeue_message(inport->channel_id, &closed);
    while (buf == NULL && !closed)
    {
	buf = yoke_recv_transient(&sender, &type, FALSE);

	if (FM_IS_SEND(type))
	{
	    ProcessSendMessage(type, buf);
	    yoke_release_transient(sender);
	}
	else if (FM_IS_CLOSE(type))
	{
	    ProcessCloseMessage(type, buf);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	}
	else
	{
	    switch(type)
	    {
	    case FM_MTYPE_OUTPORT_PERM_REQ:
		ProcessPermReq(type, buf);
		yoke_release_transient(sender);
		yoke_free_buffer(buf);
		break;

	    case FM_MTYPE_CLOSE_ACK:
		HandleCloseAck(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_SEND_PORTS:
		HandlePortSend(sender, buf);
		break;
		
	    case FM_MTYPE_INPORT_MIGRATE:
		HandleInportMigrate(sender, buf);
		yoke_free_buffer(buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
		HandleOutportMigrate(sender, buf);
		yoke_free_buffer(buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_PROCESS_RETURN:
		HandleProcessReturn(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    default:
		yoke_fatal("_fm_recv_generic(): Dunno what to do with type %s %d\n",
			   _fm_mtype_name(type), type);
	    }
	}
	
	buf = _fm_process_dequeue_message(inport->channel_id, &closed);
    }

    if (buf == NULL)
	if (closed)
	{
	    SetIOSTAT(FM_IOSTAT_EOF);
	}
	else
	    yoke_fatal("fm_recv_generic got NULL buf but channel not closed!\n");
    else
    {
	SetIOSTAT(FM_IOSTAT_OK);
    }

	    
#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("recv_generic channel_id=%d returning status=%d buf=%x\n",
		    inport->channel_id, *iostat, buf);
#endif
    return buf;
}


void FM_PROBE(f_integer *f_inport, f_integer *iostat,
	      f_integer *have_iostat, f_logical *empty)
{
    bool_t closed;
    fm_inport_t inport;

    inport = (fm_inport_t) *f_inport;

    FM_INPORT_MAGIC_CHECK(FM_PROBE, inport);

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("Probing %s\n",
		    _fm_inport_as_string(inport));
#endif

    if (inport->channel_id < 0)
    {
	if (inport->outports == (fm_outport_reference_t *) NULL)
	{
	    /* Port is unconnected */
	    SetIOSTAT(FM_IOSTAT_INPORT_UNCONNECTED);
	    *empty = FTN_LOGICAL_TRUE;
	}
	else
	{
	    /* We're locally connected and there is nothing there */
	    SetIOSTAT(FM_IOSTAT_OK);
	    *empty = FTN_LOGICAL_TRUE;
	}
    }
    else
    {
	bool_t closed;

	*empty = _fm_process_channel_is_empty(inport->channel_id, &closed);

	/*
	 * If the channel is not empty, return empty immediately.
	 *
	 * If the channel is empty and closed, return not empty immediately.
	 *
	 * If the channel is empty and not closed, try to do a
	 * receive and then return appropriately.
	 */
	
	if (!*empty)
	{
	    SetIOSTAT(FM_IOSTAT_OK);
	}
	else if (closed)
	{
	    *empty = FTN_LOGICAL_FALSE;
	    SetIOSTAT(FM_IOSTAT_OK);
	}
	else
	{
	    yoke_buffer_t buf;
	    conn_t sender;
	    int type;

	    buf = yoke_recv_transient(&sender, &type, TRUE);

	    if (buf != (yoke_buffer_t) NULL)
	    {
		if (FM_IS_SEND(type))
		{
		    ProcessSendMessage(type, buf);
		    yoke_release_transient(sender);
		}
		else if (FM_IS_CLOSE(type))
		{
		    ProcessCloseMessage(type, buf);
		    yoke_free_buffer(buf);
		    yoke_release_transient(sender);
		}
		else
		{
		    switch(type)
		    {
		    case FM_MTYPE_OUTPORT_PERM_REQ:
			ProcessPermReq(type, buf);
			yoke_release_transient(sender);
			yoke_free_buffer(buf);
			break;
			
		    case FM_MTYPE_CLOSE_ACK:
			HandleCloseAck(sender, buf);
			yoke_release_transient(sender);
			break;
			
		    case FM_MTYPE_SEND_PORTS:
			HandlePortSend(sender, buf);
			break;
			
		    case FM_MTYPE_INPORT_MIGRATE:
			HandleInportMigrate(sender, buf);
			yoke_free_buffer(buf);
			yoke_release_transient(sender);
			break;
			
		    case FM_MTYPE_OUTPORT_MIGRATE:
			HandleOutportMigrate(sender, buf);
			yoke_free_buffer(buf);
			yoke_release_transient(sender);
			break;
			
		    case FM_PROCESS_RETURN:
			HandleProcessReturn(sender, buf);
			yoke_release_transient(sender);
			break;
			
		    default:
			yoke_fatal("FM_PROBE(): Dunno what to do with type %s %d\n",
				   _fm_mtype_name(type), type);
		    }
		}
	    }
	
	    *empty = _fm_process_channel_is_empty(inport->channel_id, &closed);

	    SetIOSTAT(FM_IOSTAT_OK);

	    if (closed)
		*empty = FTN_LOGICAL_FALSE;
	}
    }
}


/*
 * _fm_handle_messages()
 *
 * This is the message notification function that is executed when
 * another processes invokes yoke_notify() with our conn_t.
 */
void _fm_handle_messages()
{
    yoke_buffer_t buf;
    conn_t sender;
    int type;
    
#ifdef DEBUG
    if (MigrateDebug(2))
    {
	yoke_printf("Received message notification\n");
    }
#endif

    /*
     * Receive Loop
     */
    
    while ((buf = yoke_recv_transient(&sender, &type, TRUE)) != NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("_fm_handle_messages(): Got message type %d %s\n",
			type, _fm_mtype_name(type));
#endif
	
	if (FM_IS_SEND(type))
	{
	    ProcessSendMessage(type, buf);
	    yoke_release_transient(sender);
	}
	else if (FM_IS_CLOSE(type))
	{
	    ProcessCloseMessage(type, buf);
	    yoke_release_transient(sender);
	}
	else
	{
	    switch(type)
	    {
	    case FM_MTYPE_OUTPORT_PERM_REQ:
		ProcessPermReq(type, buf);
		yoke_release_transient(sender);
		yoke_free_buffer(buf);
		break;
		
	    case FM_MTYPE_CLOSE_ACK:
		HandleCloseAck(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_SEND_PORTS:
		HandlePortSend(sender, buf);
		break;
		    
	    case FM_MTYPE_INPORT_MIGRATE:
		HandleInportMigrate(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
		HandleOutportMigrate(sender, buf);
		yoke_release_transient(sender);
		break;

	    case FM_PROCESS_RETURN:
		HandleProcessReturn(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    default:
		yoke_fatal("_fm_handle_messages(): Dunno what to do with type %d %s\n",
			   _fm_mtype_name(type), type);
	    }
	}
    }
}

void _fm_handle_message(conn_t sender, int type, yoke_buffer_t buf,
			bool_t process_queue)
{
    if (FM_IS_SEND(type))
    {
	ProcessSendMessage(type, buf);
	yoke_release_transient(sender);
    }
    else if (FM_IS_CLOSE(type))
    {
	ProcessCloseMessage(type, buf);
	yoke_release_transient(sender);
    }
    else
    {
	switch(type)
	{
	case FM_MTYPE_OUTPORT_PERM_REQ:
	    ProcessPermReqWithQueueFlag(type, buf, process_queue);
	    yoke_release_transient(sender);
	    yoke_free_buffer(buf);
	    break;
	    
	case FM_MTYPE_CLOSE_ACK:
	    HandleCloseAck(sender, buf);
	    yoke_release_transient(sender);
	    break;
	    
	case FM_MTYPE_SEND_PORTS:
	    HandlePortSend(sender, buf);
	    break;
	    
	case FM_MTYPE_INPORT_MIGRATE:
	    HandleInportMigrate(sender, buf);
	    yoke_release_transient(sender);
	    break;
	    
	case FM_MTYPE_OUTPORT_MIGRATE:
	    HandleOutportMigrate(sender, buf);
	    yoke_release_transient(sender);
	    break;
	    
	case FM_PROCESS_RETURN:
	    HandleProcessReturn(sender, buf);
	    yoke_release_transient(sender);
	    break;
	    
	default:
	    yoke_fatal("_fm_handle_message(): Dunno what to do with type %s %d\n",
		       _fm_mtype_name(type), type);
	}
    }
}

void _fm_invalidate_inport(fm_inport_t inport)
{
#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("Invalidating inport %s\n",
		    _fm_inport_as_string(inport));
#endif
    inport->channel_id = -1;
    inport->channel_conn = (conn_t) NULL;
    inport->outports = (fm_outport_reference_t *) NULL;
    inport->n_outports = inport->n_outports_alloc = -1;
    inport->passed_in_call = FALSE;
}

void _fm_invalidate_outport(fm_outport_t outport)
{
#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("Invalidating outport %s\n",
		    _fm_outport_as_string(outport));
#endif
    outport->channel_id = -1;
    outport->channel_conn = (conn_t) NULL;
    outport->local_inport = (fm_inport_t) NULL;
    outport->future_home = (conn_t) NULL;
    outport->inport_index = -1;
    outport->closed = FALSE;
    outport->passed_in_call = FALSE;
}

