/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "EXTERN.h"
#include "perl.h"
#undef fatal

#include "tree.h"
#include "symbol_redefines.h"

char *savestr();

extern char *strdup();

#undef USERVAR
#define USERVAR(x,y)
#undef USERSUB
#define USERSUB(enumname, name) enumname,
    
enum usersubs {
#include "usersubs.def"
    US_unused
};

#undef USERVAR
#undef USERSUB
#define USERSUB(x,y)
#define USERVAR(enumname, name) enumname,
    
enum uservars {
#include "usersubs.def"
	UV_unused
};

static int usersub();
static int userset();
static int userval();

void init_parse();

int userinit()
{
    init_parse();
}

void init_parse()
{
    char *filename = "fperl.c";
    extern void make_usub();
    struct ufuncs uf;

    uf.uf_set = userset;
    uf.uf_val = userval;

#undef USERSUB
#define USERSUB(x,y)    
#undef USERVAR
#define USERVAR(enumname, name) uf.uf_index=enumname; \
    magicname(name, &uf, sizeof(uf));

#include "usersubs.def"

#undef USERSUB
#undef USERVAR
#define USERVAR(x,y)
#define USERSUB(enumname, name) make_usub(name, enumname, usersub, filename);
#include "usersubs.def"
}

static int userset(ix, str)
int ix;
STR *str;
{
    extern int show_parse_errors;
    extern int allow_plain_common;
#ifdef YYDEBUG
    extern int yydebug;
#endif
    
    switch (ix)
    {
    case UV_show_parse_errors:
	show_parse_errors = (int) str_gnum(str);
	break;

    case UV_allow_plain_common:
	allow_plain_common = (int) str_gnum(str);
	break;

#ifdef YYDEBUG
    case UV_yydebug:
	yydebug = (int) str_gnum(str);
	break;
#endif
    }
    0;
}

static int userval(ix, str)
int ix;
STR *str;
{
    extern int show_parse_errors;
    extern int allow_plain_common;
#ifdef YYDEBUG
    extern int yydebug;
#endif
    
    switch(ix) 
    {
    case UV_show_parse_errors:
	str_numset(str, (double) show_parse_errors);
	break;

    case UV_allow_plain_common:
	str_numset(str, (double) allow_plain_common);
	break;

#ifdef YYDEBUG
    case UV_yydebug:
	str_numset(str, (double) yydebug);
	break;
#endif
    }
    0;
}

static int
usersub(ix, sp, items)
int ix;
register int sp;
register int items;
{
    STR **st = stack->ary_array + sp;
    register STR *Str;		/* used in str_get and str_gnum macros */

    switch (ix) {

    case US_parse:
	if (items != 5)
	    fatal("Usage: &parse($string, $tree, $item_type, $line, $failed_stmt)");
	else {
	    int retval;
	    char*	string =	(char*)		str_get(st[1]);
	    int item_type, failed_stmt;
	    tree_node*	tree;
	    int line = (int) str_gnum(st[4]);

	    extern int parse_line();

	    retval = parse_line(string, &tree, &item_type, line, &failed_stmt);
	    str_numset(st[0], (double) retval);
	    str_numset(st[2], (double) (int) tree);
	    str_numset(st[3], (double) item_type);
	    if (failed_stmt == -1)
		str_set(st[5], "");
	    else
		str_set(st[5], tree_type_names[failed_stmt]);
	}
	return sp;

    case US_set_input_file:
	if (items != 1)
	    fatal("Usage: &set_input_file($name)");
	else {
	    int retval;
	    char*	string =	(char*)		str_get(st[1]);

	    current_file = strdup(string);

	    str_numset(st[0], (double) 1);
	}
	return sp;

    case US_new_node:
	if (items != 1 && items != 2)
	    fatal("Usage: &new_node($type, $nitems)");
	else {
	    int retval;
	    int type = (int) str_gnum(st[1]);
	    int nitems;
	    tree_node*	tree;

	    if (items == 1)
		nitems = 3;
	    else
		nitems = (int) str_gnum(st[2]);
	    

	    tree = new_node(type, nitems);

	    str_numset(st[0], (double) (int) tree);
	}
	return sp;

    case US_add_int_kid:
	if (items != 2)
	    fatal("Usage: &add_int_kid($node, $val)");
	else {
	    int retval;
	    tree_node *	node =		(tree_node *)(int)str_gnum(st[1]);
	    int		val =		(int)		str_gnum(st[2]);

	    retval = (int) add_kid(node, new_int_node(val));
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_add_string_kid:
	if (items != 2)
	    fatal("Usage: &add_string_kid($node, $val)");
	else {
	    int retval;
	    tree_node * node =		(tree_node *)(int)str_gnum(st[1]);
	    char*	val =		(char*)		str_get(st[2]);

	    retval = (int) add_kid(node, new_string_node(val));
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_new_int_node:
	if (items != 1)
	    fatal("Usage: &new_int_node($val)");
	else {
	    int retval;
	    int		val =		(int)		str_gnum(st[1]);

	    retval = (int) new_int_node(val);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_new_string_node:
	if (items != 1)
	    fatal("Usage: &new_string_node($val)");
	else {
	    int retval;
	    char*	val =		(char*)		str_get(st[1]);

	    retval = (int) new_string_node(val);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_add_tree_kid:
	if (items != 2)
	    fatal("Usage: &add_tree_kid($node, $tree)");
	else {
	    int retval;
	    tree_node*	node =		(tree_node*)   (int) str_gnum(st[1]);
	    tree_node*	tree =		(tree_node*)	(int) str_gnum(st[2]);

	    retval = (int) add_kid(node, tree);
	    str_numset(st[0], (double) retval);
	}
	return sp;
	
    case US_add_tree_kid_to_front:
	if (items != 2)
	    fatal("Usage: &add_tree_kid_to_front($node, $tree)");
	else {
	    int retval;
	    tree_node*	node =		(tree_node*)   (int) str_gnum(st[1]);
	    tree_node*	tree =		(tree_node*)	(int) str_gnum(st[2]);

	    retval = (int) add_kid_to_front(node, tree);
	    str_numset(st[0], (double) retval);
	}
	return sp;
	
    case US_tree_num_kids:
	if (items != 1)
	    fatal("Usage: &tree_num_kids($tree)");
	else {
	    int retval;
	    tree_node*	tree =		(tree_node*)	(int) str_gnum(st[1]);

	    if (tree == NULL)
	    {
		retval = -1;
	    }
	    else
		retval = NKIDS(tree);
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tree_kid:
	if (items != 2)
	    fatal("Usage: &tree_kid($tree, $index)");
	else {
	    tree_node* retval;
	    tree_node*	tree =		(tree_node*) (int)     str_gnum(st[1]);
	    int		index =		(int)		str_gnum(st[2]);

	    if (tree == NULL)
		retval = NULL;
	    else if (index < 0 || index >= NKIDS(tree))
	    {
		printf("&tree_kid error: index out of range (0..%d)\n",
		       NKIDS(tree) - 1);
		retval = NULL;
	    }
	    else
		retval = KID(tree, index);
	    str_numset(st[0], (double) (int) retval);
	}
	return sp;

    case US_tree_set_kid:
	if (items != 3)
	    fatal("Usage: &tree_set_kid($tree, $index, $kid)");
	else {
	    tree_node* retval;
	    tree_node*	tree =		(tree_node*) (int)     str_gnum(st[1]);
	    int		index =		(int)		str_gnum(st[2]);
	    tree_node*	kid =		(tree_node*) (int)     str_gnum(st[3]);

	    if (tree == NULL)
		retval = NULL;
	    else if (index < 0 || index >= NKIDS(tree))
	    {
		printf("&tree_kid error: index out of range (0..%d)\n",
		       NKIDS(tree) - 1);
		retval = NULL;
	    }
	    else
	    {
		SET_KID(tree, index,kid);
		retval = tree;
	    }
	    str_numset(st[0], (double) (int) retval);
	}
	return sp;

    case US_tree_type:
	if (items != 1)
	    fatal("Usage: &tree_type($tree)");
	else {
	    int retval;
	    tree_node*	tree =		(tree_node*) (int) str_gnum(st[1]);

	    if (tree == NULL)
	    {
		fprintf(stderr,
			"Warning: access of null tree in tree_type\n");
		retval = -1;
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(tree);
#endif		
		retval = (int) tree->type;
	    }
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tree_label:
	if (items != 1)
	    fatal("Usage: &tree_label($tree)");
	else {
	    int retval;
	    tree_node*	tree =		(tree_node*) (int) str_gnum(st[1]);

	    if (tree == NULL)
	    {
		fprintf(stderr,
			"Warning: access of null tree in tree_type\n");
		retval = -1;
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(tree);
#endif		
		retval = (int) tree->label;
	    }
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tree_set_label:
	if (items != 2)
	    fatal("Usage: &tree_set_label($tree, $label)");
	else {
	    int retval;
	    tree_node*	tree =		(tree_node*) (int) str_gnum(st[1]);
	    int label = (int) str_gnum(st[2]);

	    if (tree == NULL)
	    {
		fprintf(stderr,
			"Warning: access of null tree in tree_type\n");
		retval = -1;
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(tree);
#endif
		tree->label = label;
		retval = (int) tree;
	    }
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tree_type_name:
	if (items != 1)
	    fatal("Usage: &tree_type_name($tree)");
	else {
	    char *retval;
	    tree_node*	tree =		(tree_node*) (int) str_gnum(st[1]);

	    if (tree == NULL)
	    {
		fprintf(stderr,
			"Warning: access of null tree in tree_type_name\n");
		retval = "";
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(tree);
#endif		
	    if (tree->type < 0 || tree->type >= LAST_TYPE)
		retval = "<<<<ERROR>>>>>";
	    else
		retval = tree_type_names[tree->type];
	    }
	    str_set(st[0], retval);
	}
	return sp;

    case US_name_of_type:
	if (items != 1)
	    fatal("Usage: &name_of_type($type)");
	else {
	    char* retval;
	    int		type =		(int)		str_gnum(st[1]);

	    if (type < 0 || type >= LAST_TYPE)
		retval = "<<<<ERROR>>>>>";
	    else
		retval = tree_type_names[type];
	    str_set(st[0], (char*) retval);
	}
	return sp;

    case US_num_of_type:
	if (items != 1)
	    fatal("Usage: &num_of_type($type)");
	else {
	    int retval;
	    char *type =		(char *)str_get(st[1]);
	    int i;

	    for (i = 0; i < LAST_TYPE; i++)
	    {
		if (strcmp(type, tree_type_names[i]) == 0)
		    break;
	    }
	    if (i == LAST_TYPE)
		retval = -1;
	    else
		retval = i;
	    
	    str_numset(st[0], (double) retval);
	}
	return sp;

    case US_tree_string_value:
	if (items != 1)
	    fatal("Usage: &tree_string_value($tree)");
	else {
	    char* retval;
	    int		tree =		(int)		str_gnum(st[1]);
	    tree_node *treep = (tree_node *) tree;

	    if (treep == NULL)
	    {
		printf("&tree_string_value error: non-string node\n");
		retval = "";
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(treep);
#endif		
	    
		if (treep->type != TREE_STRING)
		{
		    printf("&tree_string_value error: non-string node\n");
		    retval = "";
		}
		else
		    retval = treep->data.sdata;
	    }
	    str_set(st[0], (char*) retval);
	}
	return sp;

    case US_tree_int_value:
	if (items != 1)
	    fatal("Usage: &tree_int_value($tree)");
	else {
	    int retval;
	    int		tree =		(int)		str_gnum(st[1]);
	    tree_node *treep = (tree_node *) tree;

	    if (treep == NULL)
	    {
		printf("&tree_int_value error: non-int node\n");
		retval = -1;
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(treep);
#endif		
		if (treep->type != TREE_INT)
		{
		    printf("&tree_int_value error: non-int node\n");
		    retval = -1;
		}
	    }		
	    retval = treep->data.idata;

	    str_numset(st[0], (double) retval);
	}
	return sp;
    case US_tree_line_info:
	if (items != 3)
	    fatal("Usage: &tree_line_info($tree, $file, $line)");
	else {
	    int retval;
	    int		tree =		(int)		str_gnum(st[1]);
	    tree_node *treep = (tree_node *) tree;

	    if (treep == NULL)
	    {
		printf("&tree_line_info error: non-int node\n");
		retval = -1;
	    }
	    else
	    {
#ifdef PARANOID
		PARANOID_TREE_NODE_CHECK(treep);
#endif		
		str_set(st[2], treep->file);
		str_numset(st[3], (double) treep->line);
		
		retval = 1;
	    }


	    str_numset(st[0], (double) retval);
	}
	return sp;
    }
}
