/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "yoke.h" 

/*
 * Select the appropriate set of these DO_* defines to configure
 * this test program to do what you want.
 */
#define DO_FORK_TEST
/**/
#define DO_DEADLOCK_TEST
#define DO_SMALL_DEADLOCK_TEST
#define DO_BIG_DEADLOCK_TEST
#define DO_ZERO_LENGTH_TEST
/**/

#ifdef DO_FORK_TEST
static void fork_test();
#endif /* DO_FORK_TEST */
#ifdef DO_DEADLOCK_TEST
static void deadlock_test(int n_nodes, bool_t master, conn_t *conn_list);
static void swap_messages(conn_t conn, char *who);
#endif /* DO_DEADLOCK_TEST */
#ifdef DO_ZERO_LENGTH_TEST
static void zero_length_test(int n_nodes, bool_t master, conn_t *conn_list);
#endif /* DO_ZERO_LENGTH_TEST */


#ifdef FM_NEED_TRAILING_UNDERSCORES

#ifdef FM_NEED_UPPERCASE
#define YOKE_STARTUP YOKE_STARTUP_
#else  /* FM_NEED_UPPERCASE */
#define YOKE_STARTUP yoke_startup_
#endif /* FM_NEED_UPPERCASE */

#else  /* FM_NEED_TRAILING_UNDERSCORES */

#ifdef FM_NEED_UPPERCASE
#define YOKE_STARTUP YOKE_STARTUP
#else  /* FM_NEED_UPPERCASE */
#define YOKE_STARTUP yoke_startup
#endif /* FM_NEED_UPPERCASE */

#endif /* FM_NEED_TRAILING_UNDERSCORES */


extern int sleep(unsigned);

#ifdef MALLOC_DEBUG
extern int _y_malloc_debug_level;
#define YokeMallocDebug(Level) (_y_malloc_debug_level >= (Level))
#endif

#define SMALL_ARRAY_SIZE 10
#define BIG_ARRAY_SIZE   20000
static double big_array[BIG_ARRAY_SIZE];

static int sleep_time = 2;

static int handle_args(int current_arg, int arg_count)
{
    extern int atoi(char *);
    extern int printf(const char *format, ...);
    char *arg;
    
    arg = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg);

    if ((strcmp(arg, "-sleep") == 0) && ((current_arg + 1) < arg_count))
    {
	sleep_time = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,
				     current_arg + 1));
	current_arg += 2;
    }
    else
    {
	printf("handle_args(): got arg=\"%s\"\n", arg);
    }

    return current_arg;
}

void YOKE_STARTUP()
{
    int n_nodes;
    conn_t *conn_list;
    bool_t master;
    
    yoke_init("-fm", handle_args, NULL, NULL, &n_nodes, &conn_list);
    master = (n_nodes > 0);
    yoke_printf("Yoke init returns n_nodes=%d\n", n_nodes);

    yoke_printf("Process sleeping...\n");
    sleep(sleep_time);
    yoke_printf("Process continuing\n");

#ifdef DO_FORK_TEST
    fork_test();
#endif /* DO_FORK_TEST */

#ifdef DO_DEADLOCK_TEST
    deadlock_test(n_nodes, master, conn_list);
#endif /* DO_DEADLOCK_TEST */
		  
#ifdef DO_ZERO_LENGTH_TEST
    zero_length_test(n_nodes, master, conn_list);
#endif /* DO_ZERO_LENGTH_TEST */
		  
    yoke_printf("Process sleeping...\n");
    sleep(sleep_time);
    if (master)
	sleep(2);
    yoke_printf("Process exiting\n");

#ifdef MALLOC_DEBUG
    if (YokeMallocDebug(3))
	yoke_show_freed_blocks();
    yoke_show_malloc_stats();
#endif
    yoke_shutdown();
} /* yoke_startup() */

#ifdef DO_FORK_TEST
static void fork_test()
{
    int child_pid;
    conn_t child_conn;

    yoke_printf("fork_test(): parent: entering\n");

    yoke_printf("fork_test(): parent: sleeping...\n");
    sleep(sleep_time);
    yoke_printf("fork_test(): parent: forking\n");

    child_conn = yoke_fork(&child_pid);
    
    if (child_pid == 0)
    {
	/* child */
	yoke_buffer_t buf;
	conn_t c, parent_conn;
	int type;
	int value;
	
	yoke_printf("fork_test(): child: entering\n");

	yoke_printf("fork_test(): child: receiving message from parent\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	yoke_get_int(buf, &value, 1);
	yoke_get_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_get_conn(buf, &parent_conn);
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("fork_test(): child: received message: type=%d, value=%d, parent_conn=%s\n",
		    type, value, yoke_conn_as_string(parent_conn));

	
	yoke_printf("fork_test(): child: sending message 1 to parent\n");
	yoke_establish(parent_conn);
	buf = yoke_allocate_buffer(parent_conn,
				   (yoke_sizeof_int(parent_conn,1)
				    + yoke_sizeof_double(parent_conn,
							 BIG_ARRAY_SIZE)),
				   2);
	yoke_put_int(buf, &value, 1);
	yoke_put_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_send(parent_conn, buf, 5);
	yoke_release(parent_conn);
	yoke_printf("fork_test(): child: sent message 1\n");

	
	yoke_printf("fork_test(): child: sending message 2 to parent\n");
	yoke_establish(parent_conn);
	buf = yoke_allocate_buffer(parent_conn,
				   (yoke_sizeof_int(parent_conn,1)
				    + yoke_sizeof_double(parent_conn,
							 BIG_ARRAY_SIZE)),
				   2);
	yoke_put_int(buf, &value, 1);
	yoke_put_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_send(parent_conn, buf, 5);
	yoke_release(parent_conn);
	yoke_printf("fork_test(): child: sent message 2\n");

	
	yoke_printf("fork_test(): child: sleeping...\n");
	sleep(sleep_time);
	yoke_printf("fork_test(): child: exiting\n");
#ifdef MALLOC_DEBUG
	if (YokeMallocDebug(3))
	    yoke_show_freed_blocks();
	yoke_show_malloc_stats();
#endif
	yoke_shutdown();
    }
    else
    {
	/* parent */
	yoke_buffer_t buf;
	int type;
	int value = 42;
	conn_t c;
	
	yoke_printf("fork_test(): parent: forked child=%d, conn=%s\n",
		    child_pid, yoke_conn_as_string(child_conn));

	c = yoke_self();
	yoke_printf("fork_test(): parent: sending message to child with my conn=%s\n",
		    yoke_conn_as_string(c));
	yoke_establish(child_conn);
	buf = yoke_allocate_buffer(child_conn,
				   (yoke_sizeof_int(child_conn,1)
				    + yoke_sizeof_double(child_conn,
							 BIG_ARRAY_SIZE)
				    + yoke_sizeof_conn(child_conn, c)),
				   3);
	yoke_put_int(buf, &value, 1);
	yoke_put_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_put_conn(buf, c);
	yoke_send(child_conn, buf, 5);
	yoke_release(child_conn);
	yoke_printf("fork_test(): parent: sent message\n");

	
	yoke_printf("fork_test(): parent: receiving message 1 from child\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	yoke_get_int(buf, &value, 1);
	yoke_get_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("fork_test(): parent: received message 1: type=%d, value=%d\n",
		    type, value);


	yoke_printf("fork_test(): parent: receiving message 2 from child\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	yoke_get_int(buf, &value, 1);
	yoke_get_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("fork_test(): parent: received message 2: type=%d, value=%d\n",
		    type, value);
    }
    
    yoke_printf("fork_test(): parent: sleeping...\n");
    sleep(sleep_time);
    yoke_printf("fork_test(): parent: complete\n");
} /* fork_test() */
#endif /* DO_FORK_TEST */


#ifdef DO_DEADLOCK_TEST
static void deadlock_test(int n_nodes, bool_t master, conn_t *conn_list)
{
    if (master && n_nodes != 2)
    {
	yoke_fatal("deadlock_test(): Must be run on 2 or more nodes\n");
    }
	
    if (!master)
    {
	yoke_buffer_t buf;
	conn_t c, conn;
	int type;
	int value;
	
	yoke_printf("deadlock_test(): slave: entering\n");
	yoke_printf("deadlock_test(): slave: receiving master conn\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	if (type != 100)
	{
	    yoke_fatal("deadlock_test(): slave: error, expected type 100, got type %d\n", type);
	}
	yoke_get_conn(buf, &conn);
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("deadlock_test(): slave: receiving master conn=%s\n",
		    yoke_conn_as_string(conn));

	
	yoke_printf("deadlock_test(): slave: receiving big message\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	if (type != 104)
	{
	    yoke_fatal("deadlock_test(): slave: error, expected type 104, got type %d\n", type);
	}
	yoke_get_int(buf, &value, 1);
	yoke_get_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("deadlock_test(): slave: received big message\n");


	swap_messages(conn, "slave");
    }
    else
    {
	/* master */
	yoke_buffer_t buf;
	conn_t my_conn, conn;
	int value = 42;

	yoke_printf("deadlock_test(): master: entering\n");

	conn = conn_list[1];
	
	yoke_printf("deadlock_test(): master: sending on conn=%s\n",
		    yoke_conn_as_string(conn));

	/* Send my conn to the slave */
	my_conn = yoke_self();
	yoke_printf("deadlock_test(): master: sending my_conn=%s\n",
		    yoke_conn_as_string(my_conn));
	yoke_establish(conn);
	buf = yoke_allocate_buffer(conn,
				   yoke_sizeof_conn(conn, my_conn),
				   1);
	yoke_put_conn(buf, my_conn);
	yoke_send(conn, buf, 100);
	yoke_release(conn);

	
	yoke_printf("deadlock_test(): master: sending big message\n");
	yoke_establish(conn);
	buf = yoke_allocate_buffer(conn,
				   (yoke_sizeof_int(conn,1)
				    + yoke_sizeof_double(conn,
							 BIG_ARRAY_SIZE)),
				   2);
	yoke_put_int(buf, &value, 1);
	yoke_put_double(buf, big_array, BIG_ARRAY_SIZE);
	yoke_send(conn, buf, 104);
	yoke_release(conn);
	yoke_printf("deadlock_test(): master: sent big message\n");

	
	swap_messages(conn, "master");
    }
    
    yoke_printf("deadlock_test(): sleeping...\n");
    sleep(sleep_time);
    yoke_printf("deadlock_test(): complete\n");
} /* deadlock_test() */

static void swap_messages(conn_t conn, char *who)
{
    yoke_buffer_t buf;
    conn_t c;
    int type;
    int value = 42;

#ifdef DO_SMALL_DEADLOCK_TEST
    yoke_printf("deadlock_test(): %s: sending small message\n", who);
    yoke_establish(conn);
    buf = yoke_allocate_buffer(conn,
			       (yoke_sizeof_int(conn,1)
				+ yoke_sizeof_double(conn,
						     SMALL_ARRAY_SIZE)),
			       2);
    yoke_put_int(buf, &value, 1);
    yoke_put_double(buf, big_array, SMALL_ARRAY_SIZE);
    yoke_send(conn, buf, 101);
    yoke_release(conn);
    yoke_printf("deadlock_test(): %s: sent small message\n", who);
    
    
    yoke_printf("deadlock_test(): %s: receiving small message\n", who);
    buf = yoke_recv_transient(&c, &type, FALSE);
    if (type != 101)
    {
	yoke_fatal("deadlock_test(): %s: error, expected type 101, got type %d\n", who, type);
    }
    yoke_get_int(buf, &value, 1);
    yoke_get_double(buf, big_array, SMALL_ARRAY_SIZE);
    yoke_free_buffer(buf);
    yoke_release_transient(c);
    yoke_printf("deadlock_test(): %s: received small message\n", who);
#endif /* DO_SMALL_DEADLOCK_TEST */
    
    
#ifdef DO_BIG_DEADLOCK_TEST
    yoke_printf("deadlock_test(): %s: sending big message\n", who);
    yoke_establish(conn);
    buf = yoke_allocate_buffer(conn,
			       (yoke_sizeof_int(conn,1)
				+ yoke_sizeof_double(conn,
						     BIG_ARRAY_SIZE)),
			       2);
    yoke_put_int(buf, &value, 1);
    yoke_put_double(buf, big_array, BIG_ARRAY_SIZE);
    yoke_send(conn, buf, 102);
    yoke_release(conn);
    yoke_printf("deadlock_test(): %s: sent big message\n", who);
    
    
    yoke_printf("deadlock_test(): %s: receiving big message\n", who);
    buf = yoke_recv_transient(&c, &type, FALSE);
    if (type != 102)
    {
	yoke_fatal("deadlock_test(): %s: error, expected type 102, got type %d\n", who, type);
    }
    yoke_get_int(buf, &value, 1);
    yoke_get_double(buf, big_array, BIG_ARRAY_SIZE);
    yoke_free_buffer(buf);
    yoke_release_transient(c);
    yoke_printf("deadlock_test(): %s: received big message\n", who);
#endif /* DO_BIG_DEADLOCK_TEST */
    
}
#endif /* DO_DEADLOCK_TEST */


#ifdef DO_ZERO_LENGTH_TEST
static void zero_length_test(int n_nodes, bool_t master, conn_t *conn_list)
{
    if (master && n_nodes != 2)
    {
	yoke_fatal("zero_length_test(): Must be run on 2 or more nodes\n");
    }
    if (!master)
    {
	yoke_buffer_t buf;
	int type;
	conn_t c;
	
	yoke_printf("zero_length_test(): slave: entering\n");
	
	yoke_printf("zero_length_test(): slave: receiving first message\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	if (type != 200)
	{
	    yoke_fatal("zero_length_test(): slave: error, expected type 200, got type %d\n", type);
	}
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("zero_length_test(): slave: received first message\n");

	
	yoke_printf("zero_length_test(): slave: receiving second message\n");
	buf = yoke_recv_transient(&c, &type, FALSE);
	if (type != 200)
	{
	    yoke_fatal("zero_length_test(): slave: error, expected type 200, got type %d\n", type);
	}
	yoke_free_buffer(buf);
	yoke_release_transient(c);
	yoke_printf("zero_length_test(): slave: received second message\n");
    }
    else
    {
	/* master */
	yoke_buffer_t buf;
	conn_t conn;

	yoke_printf("zero_length_test(): master: entering\n");

	conn = conn_list[1];
	
	yoke_printf("zero_length_test(): master: sending on conn=%s\n",
		    yoke_conn_as_string(conn));

	/* Send my conn to the slave */
	yoke_printf("zero_length_test(): master: sending first message\n");
	yoke_establish(conn);
	buf = yoke_allocate_buffer(conn, 0, 0);
	yoke_send(conn, buf, 200);
	yoke_release(conn);
	yoke_printf("zero_length_test(): master: sent first message\n");
	
	yoke_printf("zero_length_test(): master: sending second message\n");
	yoke_establish(conn);
	buf = yoke_allocate_buffer(conn, 0, 0);
	yoke_send(conn, buf, 200);
	yoke_release(conn);
	yoke_printf("zero_length_test(): master: sent second message\n");
    }
    
    yoke_printf("zero_length_test(): sleeping...\n");
    sleep(sleep_time);
    yoke_printf("zero_length_test(): complete\n");
}
#endif /* DO_ZERO_LENGTH_TEST */
